from __future__ import annotations

import random
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest
from asgi_lifespan import LifespanManager
from litestar import Litestar, Request, Response, get
from litestar.status_codes import (
    HTTP_404_NOT_FOUND,
    HTTP_409_CONFLICT,
    HTTP_500_INTERNAL_SERVER_ERROR,
)
from litestar.testing import (
    RequestFactory,
    create_test_client,  # type: ignore
)
from litestar.types.asgi_types import HTTPResponseStartEvent
from pytest import MonkeyPatch
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session

from advanced_alchemy.exceptions import (
    DuplicateKeyError,
    ForeignKeyError,
    ImproperConfigurationError,
    IntegrityError,
    InvalidRequestError,
    NotFoundError,
    RepositoryError,
)
from advanced_alchemy.extensions.litestar._utils import set_aa_scope_state
from advanced_alchemy.extensions.litestar.exception_handler import exception_to_http_response
from advanced_alchemy.extensions.litestar.plugins import (
    SQLAlchemyAsyncConfig,
    SQLAlchemyInitPlugin,
    SQLAlchemySyncConfig,
)
from advanced_alchemy.extensions.litestar.plugins.init.config.sync import (
    autocommit_before_send_handler,
    autocommit_handler_maker,
)

if TYPE_CHECKING:
    from typing import Any, Callable

    from litestar.types import Scope


def test_default_before_send_handler() -> None:
    """Test default_before_send_handler."""

    captured_scope_state: dict[str, Any] | None = None
    config = SQLAlchemySyncConfig(connection_string="sqlite://")
    plugin = SQLAlchemyInitPlugin(config=config)

    @get()
    def test_handler(db_session: Session, scope: Scope) -> None:
        nonlocal captured_scope_state
        captured_scope_state = scope["state"]
        assert db_session is captured_scope_state[config.session_dependency_key]

    with create_test_client(route_handlers=[test_handler], plugins=[plugin]) as client:
        client.get("/")
        assert captured_scope_state is not None
        assert config.session_dependency_key not in captured_scope_state


def test_default_before_send_handle_multi() -> None:
    """Test default_before_send_handler."""

    captured_scope_state: dict[str, Any] | None = None
    config1 = SQLAlchemySyncConfig(connection_string="sqlite://")
    config2 = SQLAlchemySyncConfig(
        connection_string="sqlite://",
        session_dependency_key="other_session",
        session_scope_key="_sqlalchemy_state_2",
        engine_dependency_key="other_engine",
    )
    plugin = SQLAlchemyInitPlugin(config=[config1, config2])

    @get()
    def test_handler(db_session: Session, scope: Scope) -> None:
        nonlocal captured_scope_state
        captured_scope_state = scope["state"]
        assert db_session is captured_scope_state[config1.session_dependency_key]

    with create_test_client(route_handlers=[test_handler], plugins=[plugin]) as client:
        client.get("/")
        assert captured_scope_state is not None
        assert config1.session_dependency_key not in captured_scope_state


async def test_create_all_default(monkeypatch: MonkeyPatch) -> None:
    """Test default_before_send_handler."""

    config = SQLAlchemySyncConfig(connection_string="sqlite+aiosqlite://")
    plugin = SQLAlchemyInitPlugin(config=config)
    app = Litestar(route_handlers=[], plugins=[plugin])
    with patch.object(
        config,
        "create_all_metadata",
    ) as create_all_metadata_mock:
        async with LifespanManager(app) as _client:  # type: ignore[arg-type] # pyright: ignore[reportArgumentType]
            create_all_metadata_mock.assert_not_called()


async def test_create_all(monkeypatch: MonkeyPatch) -> None:
    """Test default_before_send_handler."""

    config = SQLAlchemySyncConfig(connection_string="sqlite+aiosqlite://", create_all=True)
    plugin = SQLAlchemyInitPlugin(config=config)
    app = Litestar(route_handlers=[], plugins=[plugin])
    with patch.object(
        config,
        "create_all_metadata",
    ) as create_all_metadata_mock:
        async with LifespanManager(app) as _client:  # type: ignore[arg-type]  # pyright: ignore[reportArgumentType]
            create_all_metadata_mock.assert_called_once()


def test_before_send_handler_success_response(create_scope: Callable[..., Scope]) -> None:
    """Test that the session is committed given a success response."""
    config = SQLAlchemySyncConfig(connection_string="sqlite://", before_send_handler=autocommit_before_send_handler)
    app = Litestar(route_handlers=[], plugins=[SQLAlchemyInitPlugin(config)])
    mock_session = MagicMock(spec=Session)
    http_scope = create_scope(app=app)
    set_aa_scope_state(http_scope, config.session_scope_key, mock_session)
    http_response_start: HTTPResponseStartEvent = {
        "type": "http.response.start",
        "status": random.randint(200, 299),
        "headers": {},
    }
    autocommit_before_send_handler(http_response_start, http_scope)
    mock_session.commit.assert_called_once()


def test_before_send_handler_success_response_autocommit(create_scope: Callable[..., Scope]) -> None:
    """Test that the session is committed given a success response."""
    config = SQLAlchemySyncConfig(connection_string="sqlite://", before_send_handler="autocommit")
    app = Litestar(route_handlers=[], plugins=[SQLAlchemyInitPlugin(config)])
    mock_session = MagicMock(spec=Session)
    http_scope = create_scope(app=app)
    set_aa_scope_state(http_scope, config.session_scope_key, mock_session)
    http_response_start: HTTPResponseStartEvent = {
        "type": "http.response.start",
        "status": random.randint(200, 299),
        "headers": {},
    }
    autocommit_before_send_handler(http_response_start, http_scope)
    mock_session.commit.assert_called_once()


def test_before_send_handler_error_response(create_scope: Callable[..., Scope]) -> None:
    """Test that the session is committed given a success response."""
    config = SQLAlchemySyncConfig(connection_string="sqlite://", before_send_handler=autocommit_before_send_handler)
    app = Litestar(route_handlers=[], plugins=[SQLAlchemyInitPlugin(config)])
    mock_session = MagicMock(spec=Session)
    http_scope = create_scope(app=app)
    set_aa_scope_state(http_scope, config.session_scope_key, mock_session)
    http_response_start: HTTPResponseStartEvent = {
        "type": "http.response.start",
        "status": random.randint(300, 599),
        "headers": {},
    }
    autocommit_before_send_handler(http_response_start, http_scope)
    mock_session.rollback.assert_called_once()


def test_autocommit_handler_maker_redirect_response(create_scope: Callable[..., Scope]) -> None:
    """Test that the handler created by the handler maker commits on redirect"""
    autocommit_redirect_handler = autocommit_handler_maker(commit_on_redirect=True)
    config = SQLAlchemySyncConfig(
        connection_string="sqlite://",
        before_send_handler=autocommit_redirect_handler,
    )
    app = Litestar(route_handlers=[], plugins=[SQLAlchemyInitPlugin(config)])
    mock_session = MagicMock(spec=Session)
    http_scope = create_scope(app=app)
    set_aa_scope_state(http_scope, config.session_scope_key, mock_session)
    http_response_start: HTTPResponseStartEvent = {
        "type": "http.response.start",
        "status": random.randint(300, 399),
        "headers": {},
    }
    autocommit_redirect_handler(http_response_start, http_scope)
    mock_session.commit.assert_called_once()


def test_autocommit_handler_maker_commit_statuses(create_scope: Callable[..., Scope]) -> None:
    """Test that the handler created by the handler maker commits on explicit statuses"""
    custom_autocommit_handler = autocommit_handler_maker(extra_commit_statuses={302, 303})
    config = SQLAlchemySyncConfig(
        connection_string="sqlite://",
        before_send_handler=custom_autocommit_handler,
    )
    app = Litestar(route_handlers=[], plugins=[SQLAlchemyInitPlugin(config)])
    mock_session = MagicMock(spec=Session)
    http_scope = create_scope(app=app)
    set_aa_scope_state(http_scope, config.session_scope_key, mock_session)
    http_response_start: HTTPResponseStartEvent = {
        "type": "http.response.start",
        "status": random.randint(302, 303),
        "headers": {},
    }
    custom_autocommit_handler(http_response_start, http_scope)
    mock_session.commit.assert_called_once()


def test_autocommit_handler_maker_rollback_statuses(create_scope: Callable[..., Scope]) -> None:
    """Test that the handler created by the handler maker rolls back on explicit statuses"""
    custom_autocommit_handler = autocommit_handler_maker(commit_on_redirect=True, extra_rollback_statuses={307, 308})
    config = SQLAlchemySyncConfig(
        connection_string="sqlite://",
        before_send_handler=custom_autocommit_handler,
    )
    app = Litestar(route_handlers=[], plugins=[SQLAlchemyInitPlugin(config)])
    mock_session = MagicMock(spec=Session)
    http_scope = create_scope(app=app)
    set_aa_scope_state(http_scope, config.session_scope_key, mock_session)
    http_response_start: HTTPResponseStartEvent = {
        "type": "http.response.start",
        "status": random.randint(307, 308),
        "headers": {},
    }
    custom_autocommit_handler(http_response_start, http_scope)
    mock_session.rollback.assert_called_once()


def test_autocommit_handler_maker_rollback_statuses_multi(create_scope: Callable[..., Scope]) -> None:
    """Test that the handler created by the handler maker rolls back on explicit statuses"""
    custom_autocommit_handler = autocommit_handler_maker(
        session_scope_key="_sqlalchemy_state_2",
        commit_on_redirect=True,
        extra_rollback_statuses={307, 308},
    )
    config1 = SQLAlchemySyncConfig(
        connection_string="sqlite://",
    )
    config2 = SQLAlchemySyncConfig(
        connection_string="sqlite://",
        before_send_handler=custom_autocommit_handler,
        session_dependency_key="other_session",
        engine_dependency_key="other_engine",
        session_scope_key="_sqlalchemy_state_2",
    )
    app = Litestar(route_handlers=[], plugins=[SQLAlchemyInitPlugin(config=[config1, config2])])
    mock_session1 = MagicMock(spec=Session)
    mock_session2 = MagicMock(spec=Session)
    http_scope = create_scope(app=app)
    set_aa_scope_state(http_scope, config1.session_scope_key, mock_session1)
    set_aa_scope_state(http_scope, config2.session_scope_key, mock_session2)
    http_response_start: HTTPResponseStartEvent = {
        "type": "http.response.start",
        "status": random.randint(307, 308),
        "headers": {},
    }
    custom_autocommit_handler(http_response_start, http_scope)
    mock_session2.rollback.assert_called_once()
    mock_session1.rollback.assert_not_called()


def test_autocommit_handler_maker_rollback_statuses_multi_bad_config(create_scope: Callable[..., Scope]) -> None:
    """Test that the handler created by the handler maker rolls back on explicit statuses"""
    with pytest.raises(ImproperConfigurationError):
        custom_autocommit_handler = autocommit_handler_maker(
            session_scope_key="_sqlalchemy_state_2",
            commit_on_redirect=True,
            extra_rollback_statuses={307, 308},
        )
        config1 = SQLAlchemySyncConfig(
            connection_string="sqlite://",
        )
        config2 = SQLAlchemySyncConfig(
            connection_string="sqlite://",
            before_send_handler=custom_autocommit_handler,
            session_dependency_key="other_session",
            session_scope_key="_sqlalchemy_state_2",
        )
        app = Litestar(route_handlers=[], plugins=[SQLAlchemyInitPlugin(config=[config1, config2])])
        mock_session1 = MagicMock(spec=Session)
        mock_session2 = MagicMock(spec=Session)
        http_scope = create_scope(app=app)
        set_aa_scope_state(http_scope, config1.session_scope_key, mock_session1)
        set_aa_scope_state(http_scope, config2.session_scope_key, mock_session2)
        http_response_start: HTTPResponseStartEvent = {
            "type": "http.response.start",
            "status": random.randint(307, 308),
            "headers": {},
        }
        custom_autocommit_handler(http_response_start, http_scope)
        mock_session2.rollback.assert_called_once()
        mock_session1.rollback.assert_not_called()


def test_autocommit_handler_maker_multi(create_scope: Callable[..., Scope]) -> None:
    """Test that the handler created by the handler maker rolls back on explicit statuses"""

    config1 = SQLAlchemySyncConfig(
        connection_string="sqlite://",
        before_send_handler="autocommit",
    )
    config2 = SQLAlchemySyncConfig(
        connection_string="sqlite://",
        before_send_handler="autocommit",
        session_dependency_key="other_session",
        engine_dependency_key="other_engine",
    )
    app = Litestar(route_handlers=[], plugins=[SQLAlchemyInitPlugin(config=[config1, config2])])
    mock_session1 = MagicMock(spec=Session)
    mock_session2 = MagicMock(spec=Session)
    http_scope = create_scope(app=app)
    set_aa_scope_state(http_scope, config1.session_scope_key, mock_session1)
    set_aa_scope_state(http_scope, config2.session_scope_key, mock_session2)
    http_response_start: HTTPResponseStartEvent = {
        "type": "http.response.start",
        "status": random.randint(307, 308),
        "headers": {},
    }
    config2.before_send_handler(http_response_start, http_scope)  # type: ignore
    mock_session2.rollback.assert_called_once()
    mock_session1.rollback.assert_not_called()


def test_autocommit_handler_maker_multi_async_and_sync(create_scope: Callable[..., Scope]) -> None:
    """Test that the handler created by the handler maker rolls back on explicit statuses"""

    config1 = SQLAlchemySyncConfig(
        connection_string="sqlite://",
        before_send_handler="autocommit",
    )
    config2 = SQLAlchemySyncConfig(
        connection_string="sqlite://",
        before_send_handler="autocommit",
        session_dependency_key="other_session",
        engine_dependency_key="other_engine",
    )
    config3 = SQLAlchemyAsyncConfig(
        connection_string="sqlite+aiosqlite://",
        before_send_handler="autocommit",
        session_dependency_key="the_session",
        engine_dependency_key="the_engine",
    )
    config4 = SQLAlchemyAsyncConfig(
        connection_string="sqlite+aiosqlite://",
        before_send_handler="autocommit",
        session_dependency_key="other_other_session",
        engine_dependency_key="other_other_engine",
    )
    app = Litestar(route_handlers=[], plugins=[SQLAlchemyInitPlugin(config=[config1, config2, config3, config4])])
    mock_session1 = MagicMock(spec=Session)
    mock_session2 = MagicMock(spec=Session)
    mock_session3 = MagicMock(spec=AsyncSession)
    mock_session4 = MagicMock(spec=AsyncSession)
    http_scope = create_scope(app=app)
    set_aa_scope_state(http_scope, config1.session_scope_key, mock_session1)
    set_aa_scope_state(http_scope, config2.session_scope_key, mock_session2)
    set_aa_scope_state(http_scope, config3.session_scope_key, mock_session3)
    set_aa_scope_state(http_scope, config4.session_scope_key, mock_session4)
    http_response_start: HTTPResponseStartEvent = {
        "type": "http.response.start",
        "status": random.randint(307, 308),
        "headers": {},
    }
    config2.before_send_handler(http_response_start, http_scope)  # type: ignore
    mock_session2.rollback.assert_called_once()
    mock_session1.rollback.assert_not_called()
    mock_session3.rollback.assert_not_called()
    mock_session4.rollback.assert_not_called()


@pytest.mark.parametrize(
    ("exc", "status"),
    [
        (IntegrityError, HTTP_409_CONFLICT),
        (ForeignKeyError, HTTP_409_CONFLICT),
        (DuplicateKeyError, HTTP_409_CONFLICT),
        (InvalidRequestError, HTTP_500_INTERNAL_SERVER_ERROR),
        (NotFoundError, HTTP_404_NOT_FOUND),
    ],
)
def test_repository_exception_to_http_response(exc: type[RepositoryError], status: int) -> None:
    """Test default exception handler."""

    config1 = SQLAlchemyAsyncConfig(connection_string="sqlite+aiosqlite://")
    config2 = SQLAlchemyAsyncConfig(
        connection_string="sqlite+aiosqlite://",
        session_dependency_key="other_session",
        session_scope_key="_sqlalchemy_state_2",
        engine_dependency_key="other_engine",
    )
    plugin = SQLAlchemyInitPlugin(config=[config1, config2])
    app = Litestar(route_handlers=[], plugins=[plugin])
    request = RequestFactory(app=app, server="testserver").get("/wherever")
    response = exception_to_http_response(request, exc())
    assert app.exception_handlers.get(exc) is None
    assert app.exception_handlers.get(RepositoryError) is not None
    assert response.status_code == status


@pytest.mark.parametrize(
    ("exc", "status"),
    [
        (IntegrityError, HTTP_409_CONFLICT),
        (ForeignKeyError, HTTP_409_CONFLICT),
        (DuplicateKeyError, HTTP_409_CONFLICT),
        (InvalidRequestError, HTTP_500_INTERNAL_SERVER_ERROR),
        (NotFoundError, HTTP_404_NOT_FOUND),
    ],
)
def test_existing_repository_exception_to_http_response(exc: type[RepositoryError], status: int) -> None:
    """Test default exception handler."""

    def handler(request: Request[Any, Any, Any], exc: RepositoryError) -> Response[Any]:
        return Response(status_code=200, content="OK")

    config1 = SQLAlchemyAsyncConfig(connection_string="sqlite+aiosqlite://")
    config2 = SQLAlchemyAsyncConfig(
        connection_string="sqlite+aiosqlite://",
        session_dependency_key="other_session",
        session_scope_key="_sqlalchemy_state_2",
        engine_dependency_key="other_engine",
    )
    plugin = SQLAlchemyInitPlugin(config=[config1, config2])
    app = Litestar(route_handlers=[], plugins=[plugin], exception_handlers={RepositoryError: handler})
    request = RequestFactory(app=app, server="testserver").get("/wherever")
    response = handler(request, exc())
    assert app.exception_handlers.get(exc) is None
    assert app.exception_handlers.get(RepositoryError) is not None
    assert app.exception_handlers.get(RepositoryError) == handler
    assert response.status_code == 200


@pytest.mark.parametrize(
    ("exc", "status"),
    [
        (IntegrityError, HTTP_409_CONFLICT),
        (ForeignKeyError, HTTP_409_CONFLICT),
        (DuplicateKeyError, HTTP_409_CONFLICT),
        (InvalidRequestError, HTTP_500_INTERNAL_SERVER_ERROR),
        (NotFoundError, HTTP_404_NOT_FOUND),
    ],
)
def test_repository_disabled_exception_to_http_response(exc: type[RepositoryError], status: int) -> None:
    """Test default exception handler."""

    config1 = SQLAlchemySyncConfig(connection_string="sqlite://", set_default_exception_handler=False)
    config2 = SQLAlchemySyncConfig(
        connection_string="sqlite://",
        session_dependency_key="other_session",
        session_scope_key="_sqlalchemy_state_2",
        engine_dependency_key="other_engine",
        set_default_exception_handler=False,
    )
    plugin = SQLAlchemyInitPlugin(config=[config1, config2])
    app = Litestar(route_handlers=[], plugins=[plugin])
    assert app.exception_handlers.get(exc) is None
    assert app.exception_handlers.get(RepositoryError) is None


def test_autocommit_handler_rollback_failure_still_closes(create_scope: Callable[..., Scope]) -> None:
    """Verify session is closed even when rollback raises in sync autocommit handler."""
    from advanced_alchemy.extensions.litestar.plugins.init.config.sync import autocommit_before_send_handler

    config = SQLAlchemySyncConfig(
        connection_string="sqlite://",
        before_send_handler=autocommit_before_send_handler,
    )
    app = Litestar(route_handlers=[], plugins=[SQLAlchemyInitPlugin(config)])
    mock_session = MagicMock(spec=Session)
    mock_session.rollback.side_effect = RuntimeError("rollback failed")
    http_scope = create_scope(app=app)
    set_aa_scope_state(http_scope, config.session_scope_key, mock_session)
    http_response_start: HTTPResponseStartEvent = {
        "type": "http.response.start",
        "status": 500,
        "headers": {},
    }
    autocommit_before_send_handler(http_response_start, http_scope)
    mock_session.rollback.assert_called_once()
    # http.response.start is a SESSION_TERMINUS event, so close runs too
    mock_session.close.assert_called_once()


def test_autocommit_handler_commit_failure_does_not_crash(create_scope: Callable[..., Scope]) -> None:
    """Verify commit failure is caught and does not crash the handler."""
    from advanced_alchemy.extensions.litestar.plugins.init.config.sync import autocommit_before_send_handler

    config = SQLAlchemySyncConfig(
        connection_string="sqlite://",
        before_send_handler=autocommit_before_send_handler,
    )
    app = Litestar(route_handlers=[], plugins=[SQLAlchemyInitPlugin(config)])
    mock_session = MagicMock(spec=Session)
    mock_session.commit.side_effect = RuntimeError("commit failed")
    http_scope = create_scope(app=app)
    set_aa_scope_state(http_scope, config.session_scope_key, mock_session)
    http_response_start: HTTPResponseStartEvent = {
        "type": "http.response.start",
        "status": 200,
        "headers": {},
    }
    # Should not raise - the exception is caught
    autocommit_before_send_handler(http_response_start, http_scope)
    mock_session.commit.assert_called_once()
    mock_session.close.assert_called_once()



"""Unit tests for service model conversion flow.

These tests verify that update input types are routed through to_model(data, "update")
and the update lifecycle hook before persistence. They also cover service-level
input normalization before repository delegation.
"""

from __future__ import annotations

import datetime
from typing import Any, Optional, cast
from unittest.mock import AsyncMock, MagicMock

import pytest
from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column

from advanced_alchemy.base import UUIDAuditBase
from advanced_alchemy.repository import SQLAlchemyAsyncRepository, SQLAlchemySyncRepository
from advanced_alchemy.repository._util import get_primary_key_info
from advanced_alchemy.repository.typing import PrimaryKeyType
from advanced_alchemy.service import SchemaDumpConfig, SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService
from advanced_alchemy.utils.serialization import ATTRS_INSTALLED, MSGSPEC_INSTALLED, PYDANTIC_INSTALLED, ModelDictT

pytestmark = [pytest.mark.unit]


class MockModel(UUIDAuditBase):
    """Minimal SQLAlchemy model for service flow tests."""

    name: Mapped[str] = mapped_column(String(length=100))  # pyright: ignore[reportUninitializedInstanceVariable]
    dob: Mapped[Optional[datetime.date]] = mapped_column(nullable=True)  # pyright: ignore[reportUninitializedInstanceVariable]


class MockRepository(SQLAlchemyAsyncRepository[MockModel]):
    """Mock repository for testing."""

    model_type = MockModel

    def __init__(self) -> None:
        # Don't call super().__init__ to avoid needing session
        self.model_type = MockModel
        self.id_attribute = "id"
        # Initialize PK info for composite PK support
        self._pk_columns, self._pk_attr_names = get_primary_key_info(MockModel)


class MockSyncRepository(SQLAlchemySyncRepository[MockModel]):
    """Mock sync repository for testing."""

    model_type = MockModel

    def __init__(self) -> None:
        # Don't call super().__init__ to avoid needing session
        self.model_type = MockModel
        self.id_attribute = "id"
        # Initialize PK info for composite PK support
        self._pk_columns, self._pk_attr_names = get_primary_key_info(MockModel)


class TrackingService(SQLAlchemyAsyncRepositoryService[MockModel, MockRepository]):
    """Service that tracks to_model() calls for testing."""

    repository_type = MockRepository

    def __init__(self) -> None:
        # Create mock repository
        self._repository = MockRepository()
        # Mock model with proper SQLAlchemy attributes
        mock_model = MockModel()
        mock_model.id = "existing-id"  # type: ignore[assignment]
        mock_model.name = "existing"
        mock_model.dob = None  # type: ignore[assignment]
        self._repository.get = AsyncMock(return_value=mock_model)  # type: ignore[method-assign]
        self._repository.update = AsyncMock(side_effect=lambda data, **kwargs: data)  # type: ignore[method-assign]

        # Track method calls
        self.to_model_calls: list[tuple[Any, Optional[str]]] = []
        self.to_model_on_update_calls: list[Any] = []

    @property
    def repository(self) -> MockRepository:
        """Return mock repository."""
        return self._repository

    async def to_model(
        self,
        data: ModelDictT[MockModel],
        operation: Optional[str] = None,
        schema_dump_config: Optional[SchemaDumpConfig] = None,
    ) -> MockModel:
        """Track to_model calls."""
        self.to_model_calls.append((data, operation))
        return await super().to_model(data, operation, schema_dump_config=schema_dump_config)

    async def to_model_on_update(self, data: ModelDictT[MockModel]) -> ModelDictT[MockModel]:
        """Track to_model_on_update calls."""
        self.to_model_on_update_calls.append(data)
        return await super().to_model_on_update(data)


class TrackingSyncService(SQLAlchemySyncRepositoryService[MockModel, MockSyncRepository]):
    """Sync service that tracks to_model() calls for testing."""

    repository_type = MockSyncRepository

    def __init__(self) -> None:
        # Create mock repository
        self._repository = MockSyncRepository()
        # Mock model with proper SQLAlchemy attributes
        mock_model = MockModel()
        mock_model.id = "existing-id"  # type: ignore[assignment]
        mock_model.name = "existing"
        mock_model.dob = None  # type: ignore[assignment]
        self._repository.get = MagicMock(return_value=mock_model)  # type: ignore[method-assign]
        self._repository.update = MagicMock(side_effect=lambda data, **kwargs: data)  # type: ignore[method-assign]

        # Track method calls
        self.to_model_calls: list[tuple[Any, Optional[str]]] = []
        self.to_model_on_update_calls: list[Any] = []

    @property
    def repository(self) -> MockSyncRepository:
        """Return mock repository."""
        return self._repository

    def to_model(
        self,
        data: ModelDictT[MockModel],
        operation: Optional[str] = None,
        schema_dump_config: Optional[SchemaDumpConfig] = None,
    ) -> MockModel:
        """Track to_model calls."""
        self.to_model_calls.append((data, operation))
        return super().to_model(data, operation, schema_dump_config=schema_dump_config)

    def to_model_on_update(self, data: ModelDictT[MockModel]) -> ModelDictT[MockModel]:
        """Track to_model_on_update calls."""
        self.to_model_on_update_calls.append(data)
        return super().to_model_on_update(data)


# Tests for async service


@pytest.mark.asyncio
async def test_delete_many_passes_id_only_input_through_without_copying() -> None:
    """Id-only delete_many() input should keep the existing bulk-delete fast path."""
    service = TrackingService()
    captured: dict[str, Any] = {}

    async def delete_many(item_ids: list[Any], **_: Any) -> list[MockModel]:
        captured["item_ids"] = item_ids
        return []

    service.repository.delete_many = AsyncMock(side_effect=delete_many)  # type: ignore[method-assign]

    item_ids: list[PrimaryKeyType] = ["first-id", "second-id"]
    await service.delete_many(item_ids)

    assert captured["item_ids"] is item_ids


@pytest.mark.asyncio
async def test_update_dict_calls_to_model_with_operation() -> None:
    """Test that update() with dict data calls to_model(data, 'update')."""
    service = TrackingService()

    # Update with dict data and item_id
    await service.update({"name": "Updated Name"}, item_id="test-id")

    # Verify to_model was called with operation="update"
    assert len(service.to_model_calls) == 1
    data, operation = service.to_model_calls[0]
    assert operation == "update"
    assert data == {"name": "Updated Name"}

    # Verify to_model_on_update was also called (via operation_map)
    assert len(service.to_model_on_update_calls) == 1


@pytest.mark.asyncio
async def test_update_model_instance_calls_to_model_with_operation() -> None:
    """Test that update() with model instance calls to_model(data, 'update')."""
    service = TrackingService()

    # Update with model instance
    model = MockModel()
    model.id = "test-id"  # type: ignore[assignment]
    model.name = "Updated Name"
    await service.update(model)

    # Verify to_model was called with operation="update"
    assert len(service.to_model_calls) == 1
    data, operation = service.to_model_calls[0]
    assert operation == "update"
    assert data is model


@pytest.mark.skipif(not PYDANTIC_INSTALLED, reason="Pydantic not installed")
@pytest.mark.asyncio
async def test_update_pydantic_calls_to_model_with_operation() -> None:
    """Test that update() with Pydantic data calls to_model(data, 'update')."""
    from pydantic import BaseModel

    class AuthorSchema(BaseModel):
        name: str

    service = TrackingService()

    # Update with Pydantic model
    schema = AuthorSchema(name="Updated Name")
    await service.update(schema, item_id="test-id")

    # Verify to_model was called with operation="update"
    assert len(service.to_model_calls) == 1
    data, operation = service.to_model_calls[0]
    assert operation == "update"
    assert isinstance(data, AuthorSchema)


@pytest.mark.skipif(not PYDANTIC_INSTALLED, reason="Pydantic not installed")
@pytest.mark.asyncio
async def test_update_pydantic_per_call_schema_dump_config_includes_unset_defaults() -> None:
    """Per-call schema_dump_config should include Pydantic defaults when exclude_unset is disabled."""
    from pydantic import BaseModel

    default_dob = datetime.date(2000, 1, 1)

    class AuthorSchema(BaseModel):
        name: str
        dob: datetime.date = default_dob

    service = TrackingService()

    result = await service.update(
        AuthorSchema(name="Updated Name"),
        item_id="test-id",
        schema_dump_config=SchemaDumpConfig(exclude_unset=False),
    )

    assert result.name == "Updated Name"
    assert result.dob == default_dob


@pytest.mark.skipif(not PYDANTIC_INSTALLED, reason="Pydantic not installed")
@pytest.mark.asyncio
async def test_update_pydantic_class_schema_dump_config_includes_unset_defaults() -> None:
    """Service class schema_dump_config should apply when no per-call override is passed."""
    from pydantic import BaseModel

    default_dob = datetime.date(2000, 1, 1)

    class AuthorSchema(BaseModel):
        name: str
        dob: datetime.date = default_dob

    class IncludeUnsetDefaultsService(TrackingService):
        schema_dump_config = SchemaDumpConfig(exclude_unset=False)

    service = IncludeUnsetDefaultsService()

    result = await service.update(AuthorSchema(name="Updated Name"), item_id="test-id")

    assert result.name == "Updated Name"
    assert result.dob == default_dob


@pytest.mark.skipif(not PYDANTIC_INSTALLED, reason="Pydantic not installed")
@pytest.mark.asyncio
async def test_update_pydantic_per_call_schema_dump_config_does_not_leak() -> None:
    """A per-call schema_dump_config override should not affect later service calls."""
    from pydantic import BaseModel

    default_dob = datetime.date(2000, 1, 1)

    class AuthorSchema(BaseModel):
        name: str
        dob: datetime.date = default_dob

    service = TrackingService()
    first_existing = MockModel()
    first_existing.id = "first-id"  # type: ignore[assignment]
    first_existing.name = "first"
    first_existing.dob = None  # type: ignore[assignment]
    second_existing = MockModel()
    second_existing.id = "second-id"  # type: ignore[assignment]
    second_existing.name = "second"
    second_existing.dob = None  # type: ignore[assignment]
    service.repository.get = AsyncMock(side_effect=[first_existing, second_existing])  # type: ignore[method-assign]

    first_result = await service.update(
        AuthorSchema(name="First Updated"),
        item_id="first-id",
        schema_dump_config=SchemaDumpConfig(exclude_unset=False),
    )
    second_result = await service.update(AuthorSchema(name="Second Updated"), item_id="second-id")

    assert first_result.dob == default_dob
    assert second_result.name == "Second Updated"
    assert second_result.dob is None


@pytest.mark.skipif(not MSGSPEC_INSTALLED, reason="msgspec not installed")
@pytest.mark.asyncio
async def test_update_msgspec_calls_to_model_with_operation() -> None:
    """Test that update() with msgspec data calls to_model(data, 'update')."""
    import msgspec

    class AuthorStruct(msgspec.Struct):
        name: str

    service = TrackingService()

    # Update with msgspec struct
    struct = AuthorStruct(name="Updated Name")
    await service.update(struct, item_id="test-id")

    # Verify to_model was called with operation="update"
    assert len(service.to_model_calls) == 1
    data, operation = service.to_model_calls[0]
    assert operation == "update"
    assert isinstance(data, AuthorStruct)


@pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
@pytest.mark.asyncio
async def test_update_attrs_calls_to_model_with_operation() -> None:
    """Test that update() with attrs data calls to_model(data, 'update')."""
    from attrs import define

    @define
    class AuthorAttrs:
        name: str

    service = TrackingService()

    # Update with attrs instance
    attrs_obj = AuthorAttrs(name="Updated Name")
    await service.update(attrs_obj, item_id="test-id")

    # Verify to_model was called with operation="update"
    assert len(service.to_model_calls) == 1
    data, operation = service.to_model_calls[0]
    assert operation == "update"
    assert isinstance(data, AuthorAttrs)


@pytest.mark.asyncio
async def test_update_operation_map_routes_to_to_model_on_update() -> None:
    """Test that to_model() with operation='update' routes to to_model_on_update()."""
    service = TrackingService()

    # Update with dict data
    await service.update({"name": "updated"}, item_id="test-id")

    # Verify both methods were called
    assert len(service.to_model_calls) == 1
    assert len(service.to_model_on_update_calls) == 1

    # Verify operation_map routing worked
    _, operation = service.to_model_calls[0]
    assert operation == "update"


@pytest.mark.asyncio
async def test_update_propagates_with_for_update_flag() -> None:
    """Ensure the async service forwards locking hints to the repository."""

    service = TrackingService()
    await service.update({"name": "updated"}, item_id="test-id", with_for_update=True)

    get_mock = cast(AsyncMock, service.repository.get)
    get_mock.assert_awaited_once()
    assert get_mock.call_args.kwargs["with_for_update"] is True


# Tests for sync service


def test_sync_delete_many_passes_id_only_input_through_without_copying() -> None:
    """Sync id-only delete_many() input should keep the existing bulk-delete fast path."""
    service = TrackingSyncService()
    captured: dict[str, Any] = {}

    def delete_many(item_ids: list[Any], **_: Any) -> list[MockModel]:
        captured["item_ids"] = item_ids
        return []

    service.repository.delete_many = MagicMock(side_effect=delete_many)  # type: ignore[method-assign]

    item_ids: list[PrimaryKeyType] = ["first-id", "second-id"]
    service.delete_many(item_ids)

    assert captured["item_ids"] is item_ids


def test_sync_delete_many_resolves_model_instances_before_delegating() -> None:
    """Sync delete_many() should resolve model instances to primary key values."""
    service = TrackingSyncService()
    captured: dict[str, Any] = {}

    def delete_many(item_ids: list[Any], **_: Any) -> list[MockModel]:
        captured["item_ids"] = item_ids
        return []

    service.repository.delete_many = MagicMock(side_effect=delete_many)  # type: ignore[method-assign]

    model = MockModel()
    model.id = "model-id"  # type: ignore[assignment]
    item_ids = cast("list[PrimaryKeyType]", [model, "raw-id"])
    service.delete_many(item_ids)

    assert captured["item_ids"] == ["model-id", "raw-id"]
    assert captured["item_ids"] is not item_ids


def test_sync_update_dict_calls_to_model_with_operation() -> None:
    """Test that sync update() with dict data calls to_model(data, 'update')."""
    service = TrackingSyncService()

    # Update with dict data and item_id
    service.update({"name": "Updated Name"}, item_id="test-id")

    # Verify to_model was called with operation="update"
    assert len(service.to_model_calls) == 1
    data, operation = service.to_model_calls[0]
    assert operation == "update"
    assert data == {"name": "Updated Name"}

    # Verify to_model_on_update was also called (via operation_map)
    assert len(service.to_model_on_update_calls) == 1


def test_sync_update_model_instance_calls_to_model_with_operation() -> None:
    """Test that sync update() with model instance calls to_model(data, 'update')."""
    service = TrackingSyncService()

    # Update with model instance
    model = MockModel()
    model.id = "test-id"  # type: ignore[assignment]
    model.name = "Updated Name"
    service.update(model)

    # Verify to_model was called with operation="update"
    assert len(service.to_model_calls) == 1
    data, operation = service.to_model_calls[0]
    assert operation == "update"
    assert data is model


@pytest.mark.skipif(not PYDANTIC_INSTALLED, reason="Pydantic not installed")
def test_sync_update_pydantic_per_call_schema_dump_config_includes_unset_defaults() -> None:
    """Sync services should honor per-call schema_dump_config for Pydantic defaults."""
    from pydantic import BaseModel

    default_dob = datetime.date(2000, 1, 1)

    class AuthorSchema(BaseModel):
        name: str
        dob: datetime.date = default_dob

    service = TrackingSyncService()

    result = service.update(
        AuthorSchema(name="Updated Name"),
        item_id="test-id",
        schema_dump_config=SchemaDumpConfig(exclude_unset=False),
    )

    assert result.name == "Updated Name"
    assert result.dob == default_dob


def test_sync_update_propagates_with_for_update_flag() -> None:
    """Ensure the sync service forwards the locking flag to its repository."""

    service = TrackingSyncService()
    service.update({"name": "updated"}, item_id="test-id", with_for_update=True)

    get_mock = cast(MagicMock, service.repository.get)
    get_mock.assert_called_once()
    assert get_mock.call_args.kwargs["with_for_update"] is True


# Tests for update lifecycle overrides


@pytest.mark.asyncio
async def test_update_invokes_overridden_to_model_on_update() -> None:
    """An overridden update lifecycle hook should be invoked during update conversion."""

    class UpdateHookService(SQLAlchemyAsyncRepositoryService[MockModel, MockRepository]):
        repository_type = MockRepository

        def __init__(self) -> None:
            self._repository = MockRepository()
            mock_model = MockModel()
            mock_model.id = "test-id"  # type: ignore[assignment]
            mock_model.name = "Old Name"
            self._repository.get = AsyncMock(return_value=mock_model)  # type: ignore[method-assign]
            self._repository.update = AsyncMock(side_effect=lambda data, **kwargs: data)  # type: ignore[method-assign]
            self.update_hook_called = False

        @property
        def repository(self) -> MockRepository:
            return self._repository

        async def to_model_on_update(self, data: ModelDictT[MockModel]) -> ModelDictT[MockModel]:
            self.update_hook_called = True
            return await super().to_model_on_update(data)

    service = UpdateHookService()
    await service.update({"name": "Updated Name"}, item_id="test-id")

    assert service.update_hook_called


# Edge case tests


@pytest.mark.asyncio
async def test_update_without_item_id_uses_model_id() -> None:
    """Test update without item_id uses ID from model instance."""
    service = TrackingService()

    # Update with model that has ID
    model = MockModel()
    model.id = "model-id"  # type: ignore[assignment]
    model.name = "Updated Name"
    await service.update(model)

    # Should work - uses model's ID
    assert len(service.to_model_calls) == 1


@pytest.mark.asyncio
async def test_update_preserves_existing_instance_attributes() -> None:
    """Test that update with item_id preserves existing instance attributes."""
    service = TrackingService()

    # Update only name field
    result = await service.update({"name": "New Name"}, item_id="test-id")

    # Existing ID should be preserved from existing instance
    assert result.name == "New Name"
    assert result.id == "existing-id"  # From mock repository's get()


def test_sync_update_preserves_existing_instance_attributes() -> None:
    """Test that sync update with item_id preserves existing instance attributes."""
    service = TrackingSyncService()

    # Update only name field
    result = service.update({"name": "New Name"}, item_id="test-id")

    # Existing ID should be preserved from existing instance
    assert result.name == "New Name"
    assert result.id == "existing-id"  # From mock repository's get()



# /// script
# dependencies = [
#   "advanced_alchemy",
#   "fastapi[standard]",
#   "orjson"
# ]
# ///
import datetime
from typing import Annotated, Optional
from uuid import UUID

from fastapi import APIRouter, Depends, FastAPI
from pydantic import BaseModel
from sqlalchemy.orm import Mapped

from advanced_alchemy.extensions.fastapi import (
    AdvancedAlchemy,
    AsyncSessionConfig,
    SQLAlchemyAsyncConfig,
    base,
    filters,
    repository,
    service,
)

alchemy_config = SQLAlchemyAsyncConfig(
    connection_string="sqlite+aiosqlite:///test.sqlite",
    session_config=AsyncSessionConfig(expire_on_commit=False),
    create_all=True,
)
app = FastAPI()
alchemy = AdvancedAlchemy(config=alchemy_config, app=app)
author_router = APIRouter()


class AuthorModel(base.UUIDBase):
    __tablename__ = "author"
    name: Mapped[str]
    dob: Mapped[Optional[datetime.date]]


class Author(BaseModel):
    id: Optional[UUID]
    name: str
    dob: Optional[datetime.date]


class AuthorService(service.SQLAlchemyAsyncRepositoryService[AuthorModel]):
    """Author repository."""

    class Repo(repository.SQLAlchemyAsyncRepository[AuthorModel]):
        """Author repository."""

        model_type = AuthorModel

    repository_type = Repo


@author_router.get(path="/authors", response_model=service.OffsetPagination[Author])
async def list_authors(
    authors_service: Annotated[AuthorService, Depends(alchemy.provide_service(AuthorService))],
    filters: Annotated[
        list[filters.FilterTypes],
        Depends(
            alchemy.provide_filters(
                {
                    "id_filter": UUID,
                    "pagination_type": "limit_offset",
                    "search": "name",
                    "search_ignore_case": True,
                }
            )
        ),
    ],
) -> service.OffsetPagination[AuthorModel]:
    results, total = await authors_service.get_many_and_count(*filters)
    return authors_service.to_schema(results, total, filters=filters)


app.include_router(author_router)


if __name__ == "__main__":
    """Launches the FastAPI CLI with the database commands registered
    Run `uv run examples/fastapi/fastapi_service.py` to launch the FastAPI CLI with the database commands registered
    """
    from fastapi_cli.cli import app as fastapi_cli_app  # pyright: ignore[reportUnknownVariableType]

    from advanced_alchemy.extensions.fastapi import assign_cli_group

    assign_cli_group(app)
    fastapi_cli_app()



"""Tests for advanced_alchemy.utils.fixtures module."""

import csv
import gzip
import io
import json
import tempfile
import zipfile
from collections.abc import Generator
from pathlib import Path
from typing import Any

import pytest

from advanced_alchemy.utils.fixtures import open_fixture, open_fixture_async


@pytest.fixture
def sample_data() -> "list[dict[str, Any]]":
    """Sample JSON data for testing."""
    return [
        {"id": 1, "name": "Alice", "email": "alice@example.com"},
        {"id": 2, "name": "Bob", "email": "bob@example.com"},
        {"id": 3, "name": "Charlie", "email": "charlie@example.com"},
    ]


@pytest.fixture
def sample_csv_data() -> "list[dict[str, str]]":
    """Sample CSV data for testing (note: CSV values are strings)."""
    return [
        {"id": "1", "name": "Alice", "email": "alice@example.com"},
        {"id": "2", "name": "Bob", "email": "bob@example.com"},
        {"id": "3", "name": "Charlie", "email": "charlie@example.com"},
    ]


@pytest.fixture
def temp_fixtures_dir(
    sample_data: "list[dict[str, Any]]", sample_csv_data: "list[dict[str, str]]"
) -> "Generator[Path, None, None]":
    """Create temporary directory with test fixtures in various formats."""
    with tempfile.TemporaryDirectory() as temp_dir:
        fixtures_path = Path(temp_dir)

        # Create plain JSON fixture
        json_file = fixtures_path / "users.json"
        with open(json_file, "w", encoding="utf-8") as f:
            json.dump(sample_data, f, indent=2)

        # Create gzipped JSON fixture
        gz_file = fixtures_path / "users_gz.json.gz"
        with gzip.open(gz_file, "wt", encoding="utf-8") as f:
            json.dump(sample_data, f, indent=2)

        # Create zipped JSON fixture (single file)
        zip_file = fixtures_path / "users_zip.json.zip"
        with zipfile.ZipFile(zip_file, "w", zipfile.ZIP_DEFLATED) as zf:
            zf.writestr("users_zip.json", json.dumps(sample_data, indent=2))

        # Create zipped JSON fixture with multiple files (should pick the first)
        multi_zip_file = fixtures_path / "users_multi.json.zip"
        with zipfile.ZipFile(multi_zip_file, "w", zipfile.ZIP_DEFLATED) as zf:
            zf.writestr("other.json", json.dumps([{"other": "data"}]))
            zf.writestr("users_multi.json", json.dumps(sample_data, indent=2))

        # Create zipped JSON fixture with preferred name match
        preferred_zip_file = fixtures_path / "users_preferred.json.zip"
        with zipfile.ZipFile(preferred_zip_file, "w", zipfile.ZIP_DEFLATED) as zf:
            zf.writestr("other.json", json.dumps([{"other": "data"}]))
            zf.writestr("users_preferred.json", json.dumps(sample_data, indent=2))

        # Create empty zip file (should raise error)
        empty_zip_file = fixtures_path / "empty.json.zip"
        with zipfile.ZipFile(empty_zip_file, "w", zipfile.ZIP_DEFLATED):
            pass  # Empty zip

        # Create zip with no JSON files
        no_json_zip_file = fixtures_path / "no_json.json.zip"
        with zipfile.ZipFile(no_json_zip_file, "w", zipfile.ZIP_DEFLATED) as zf:
            zf.writestr("readme.txt", "No JSON files here")

        # Create plain CSV fixture
        csv_file = fixtures_path / "users_csv.csv"
        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            if sample_csv_data:
                writer = csv.DictWriter(f, fieldnames=sample_csv_data[0].keys())
                writer.writeheader()
                writer.writerows(sample_csv_data)

        # Create gzipped CSV fixture
        gz_csv_file = fixtures_path / "users_csv_gz.csv.gz"
        with gzip.open(gz_csv_file, "wt", encoding="utf-8", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=sample_csv_data[0].keys())
            writer.writeheader()
            writer.writerows(sample_csv_data)

        # Create zipped CSV fixture (single file)
        zip_csv_file = fixtures_path / "users_csv_zip.csv.zip"
        with zipfile.ZipFile(zip_csv_file, "w", zipfile.ZIP_DEFLATED) as zf:
            csv_content = io.StringIO()
            writer = csv.DictWriter(csv_content, fieldnames=sample_csv_data[0].keys())
            writer.writeheader()
            writer.writerows(sample_csv_data)
            zf.writestr("users_csv_zip.csv", csv_content.getvalue())

        # Create zipped CSV fixture with multiple files (should pick matching name)
        multi_zip_csv_file = fixtures_path / "users_csv_multi.csv.zip"
        with zipfile.ZipFile(multi_zip_csv_file, "w", zipfile.ZIP_DEFLATED) as zf:
            # Add a different CSV first
            other_content = io.StringIO()
            other_writer = csv.DictWriter(other_content, fieldnames=["other"])
            other_writer.writeheader()
            other_writer.writerow({"other": "data"})
            zf.writestr("other.csv", other_content.getvalue())

            # Add the actual fixture
            csv_content = io.StringIO()
            writer = csv.DictWriter(csv_content, fieldnames=sample_csv_data[0].keys())
            writer.writeheader()
            writer.writerows(sample_csv_data)
            zf.writestr("users_csv_multi.csv", csv_content.getvalue())

        # Create empty CSV zip file (should raise error)
        empty_csv_zip = fixtures_path / "empty_csv.csv.zip"
        with zipfile.ZipFile(empty_csv_zip, "w", zipfile.ZIP_DEFLATED):
            pass

        # Create zip with no CSV files
        no_csv_zip = fixtures_path / "no_csv.csv.zip"
        with zipfile.ZipFile(no_csv_zip, "w", zipfile.ZIP_DEFLATED) as zf:
            zf.writestr("readme.txt", "No CSV files here")

        yield fixtures_path


class TestOpenFixture:
    """Test cases for synchronous open_fixture function."""

    def test_open_plain_json_fixture(self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]") -> None:
        """Test loading plain JSON fixture."""
        result = open_fixture(temp_fixtures_dir, "users")
        assert result == sample_data

    def test_open_gzipped_fixture(self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]") -> None:
        """Test loading gzipped JSON fixture."""
        result = open_fixture(temp_fixtures_dir, "users_gz")
        assert result == sample_data

    def test_open_zipped_fixture(self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]") -> None:
        """Test loading zipped JSON fixture."""
        result = open_fixture(temp_fixtures_dir, "users_zip")
        assert result == sample_data

    def test_open_zipped_fixture_multiple_files(
        self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]"
    ) -> None:
        """Test loading zipped JSON fixture with multiple files, should prefer matching name."""
        result = open_fixture(temp_fixtures_dir, "users_multi")
        assert result == sample_data

    def test_open_zipped_fixture_preferred_name(
        self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]"
    ) -> None:
        """Test loading zipped JSON fixture prefers file with matching name."""
        result = open_fixture(temp_fixtures_dir, "users_preferred")
        assert result == sample_data

    def test_case_insensitive_support(self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]") -> None:
        """Test case-insensitive fixture loading (uppercase takes priority for compressed files)."""
        # Create uppercase gzipped file
        uppercase_gz_file = temp_fixtures_dir / "TESTCASE.json.gz"
        with gzip.open(uppercase_gz_file, "wt", encoding="utf-8") as f:
            json.dump(sample_data, f)

        # Test that uppercase is found
        result = open_fixture(temp_fixtures_dir, "testcase")
        assert result == sample_data

        # Create lowercase version and test priority (uppercase should still win)
        lowercase_gz_file = temp_fixtures_dir / "testcase.json.gz"
        lowercase_data = [{"different": "data"}]
        with gzip.open(lowercase_gz_file, "wt", encoding="utf-8") as f:
            json.dump(lowercase_data, f)

        # Should still load uppercase version first
        result = open_fixture(temp_fixtures_dir, "testcase")
        assert result == sample_data  # Original data, not lowercase_data

        # Remove uppercase, should fallback to lowercase
        uppercase_gz_file.unlink()
        result = open_fixture(temp_fixtures_dir, "testcase")
        assert result == lowercase_data

    def test_file_format_priority(self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]") -> None:
        """Test that plain JSON is preferred over compressed formats."""
        # Create all three formats for the same fixture name
        json_file = temp_fixtures_dir / "priority.json"
        gz_file = temp_fixtures_dir / "priority.json.gz"
        zip_file = temp_fixtures_dir / "priority.json.zip"

        # Different data for each format to test which one is loaded
        plain_data = [{"format": "plain"}]
        gz_data = [{"format": "gzip"}]
        zip_data = [{"format": "zip"}]

        with open(json_file, "w", encoding="utf-8") as f:
            json.dump(plain_data, f)

        with gzip.open(gz_file, "wt", encoding="utf-8") as f:
            json.dump(gz_data, f)

        with zipfile.ZipFile(zip_file, "w", zipfile.ZIP_DEFLATED) as zf:
            zf.writestr("priority.json", json.dumps(zip_data))

        # Should load plain JSON first
        result = open_fixture(temp_fixtures_dir, "priority")
        assert result == plain_data

        # Remove plain JSON, should load gzip
        json_file.unlink()
        result = open_fixture(temp_fixtures_dir, "priority")
        assert result == gz_data

        # Remove gzip, should load zip
        gz_file.unlink()
        result = open_fixture(temp_fixtures_dir, "priority")
        assert result == zip_data

    def test_fixture_not_found(self, temp_fixtures_dir: Path) -> None:
        """Test FileNotFoundError when fixture doesn't exist."""
        with pytest.raises(FileNotFoundError) as exc_info:
            open_fixture(temp_fixtures_dir, "nonexistent")

        assert "Could not find the nonexistent fixture" in str(exc_info.value)
        assert "(tried .json, .json.gz, .json.zip, .csv, .csv.gz, .csv.zip with case variations)" in str(exc_info.value)

    def test_empty_zip_file(self, temp_fixtures_dir: Path) -> None:
        """Test error handling for empty zip file."""
        with pytest.raises(ValueError) as exc_info:
            open_fixture(temp_fixtures_dir, "empty")

        assert "No JSON files found in zip archive" in str(exc_info.value)

    def test_zip_with_no_json_files(self, temp_fixtures_dir: Path) -> None:
        """Test error handling for zip file with no JSON files."""
        with pytest.raises(ValueError) as exc_info:
            open_fixture(temp_fixtures_dir, "no_json")

        assert "No JSON files found in zip archive" in str(exc_info.value)

    def test_corrupted_gzip_file(self, temp_fixtures_dir: Path) -> None:
        """Test error handling for corrupted gzip file."""
        # Create corrupted gzip file
        corrupted_file = temp_fixtures_dir / "corrupted.json.gz"
        with open(corrupted_file, "wb") as f:
            f.write(b"not a gzip file")

        with pytest.raises(OSError) as exc_info:
            open_fixture(temp_fixtures_dir, "corrupted")

        assert "Error reading fixture file" in str(exc_info.value)

    def test_corrupted_zip_file(self, temp_fixtures_dir: Path) -> None:
        """Test error handling for corrupted zip file."""
        # Create corrupted zip file
        corrupted_file = temp_fixtures_dir / "corrupted_zip.json.zip"
        with open(corrupted_file, "wb") as f:
            f.write(b"not a zip file")

        with pytest.raises(OSError) as exc_info:
            open_fixture(temp_fixtures_dir, "corrupted_zip")

        assert "Error reading fixture file" in str(exc_info.value)

    def test_invalid_json_content(self, temp_fixtures_dir: Path) -> None:
        """Test error handling for invalid JSON content."""
        invalid_file = temp_fixtures_dir / "invalid.json"
        with open(invalid_file, "w", encoding="utf-8") as f:
            f.write("{ invalid json content")

        with pytest.raises(Exception):  # decode_json will raise an appropriate exception
            open_fixture(temp_fixtures_dir, "invalid")


class TestOpenFixtureCSV:
    """Test cases for CSV fixture loading."""

    def test_open_plain_csv_fixture(self, temp_fixtures_dir: Path, sample_csv_data: "list[dict[str, str]]") -> None:
        """Test loading plain CSV fixture."""
        result = open_fixture(temp_fixtures_dir, "users_csv")
        assert result == sample_csv_data
        # Verify it's a list of dicts
        assert isinstance(result, list)
        assert all(isinstance(row, dict) for row in result)

    def test_open_gzipped_csv_fixture(self, temp_fixtures_dir: Path, sample_csv_data: "list[dict[str, str]]") -> None:
        """Test loading gzipped CSV fixture."""
        result = open_fixture(temp_fixtures_dir, "users_csv_gz")
        assert result == sample_csv_data

    def test_open_zipped_csv_fixture(self, temp_fixtures_dir: Path, sample_csv_data: "list[dict[str, str]]") -> None:
        """Test loading zipped CSV fixture."""
        result = open_fixture(temp_fixtures_dir, "users_csv_zip")
        assert result == sample_csv_data

    def test_open_zipped_csv_fixture_multiple_files(
        self, temp_fixtures_dir: Path, sample_csv_data: "list[dict[str, str]]"
    ) -> None:
        """Test loading zipped CSV fixture with multiple files, should prefer matching name."""
        result = open_fixture(temp_fixtures_dir, "users_csv_multi")
        assert result == sample_csv_data

    def test_csv_values_are_strings(self, temp_fixtures_dir: Path) -> None:
        """Test that CSV values are strings (important difference from JSON)."""
        result = open_fixture(temp_fixtures_dir, "users_csv")
        # CSV returns strings, not integers
        assert result[0]["id"] == "1"
        assert isinstance(result[0]["id"], str)

    def test_json_priority_over_csv(
        self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]", sample_csv_data: "list[dict[str, str]]"
    ) -> None:
        """Test that JSON fixtures are loaded before CSV when both exist."""
        # Create both JSON and CSV fixtures with same name
        json_file = temp_fixtures_dir / "priority_format.json"
        csv_file = temp_fixtures_dir / "priority_format.csv"

        with open(json_file, "w", encoding="utf-8") as f:
            json.dump(sample_data, f)

        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=sample_csv_data[0].keys())
            writer.writeheader()
            writer.writerows(sample_csv_data)

        # Should load JSON first
        result = open_fixture(temp_fixtures_dir, "priority_format")
        assert result == sample_data  # JSON data, not CSV data

        # Remove JSON, should load CSV
        json_file.unlink()
        result = open_fixture(temp_fixtures_dir, "priority_format")
        assert result == sample_csv_data

    def test_empty_csv_zip_file(self, temp_fixtures_dir: Path) -> None:
        """Test error handling for empty CSV zip file."""
        with pytest.raises(ValueError) as exc_info:
            open_fixture(temp_fixtures_dir, "empty_csv")
        assert "No CSV files found in zip archive" in str(exc_info.value)

    def test_csv_zip_with_no_csv_files(self, temp_fixtures_dir: Path) -> None:
        """Test error handling for zip file with no CSV files."""
        with pytest.raises(ValueError) as exc_info:
            open_fixture(temp_fixtures_dir, "no_csv")
        assert "No CSV files found in zip archive" in str(exc_info.value)

    def test_csv_with_special_characters(self, temp_fixtures_dir: Path) -> None:
        """Test CSV with special characters, quotes, and commas."""
        special_data = [
            {"name": "O'Brien", "description": "Has apostrophe"},
            {"name": "Smith, Jr.", "description": "Has comma"},
            {"name": 'Quote"Test', "description": "Has quote"},
        ]

        csv_file = temp_fixtures_dir / "special.csv"
        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=special_data[0].keys())
            writer.writeheader()
            writer.writerows(special_data)

        result = open_fixture(temp_fixtures_dir, "special")
        assert result == special_data

    def test_csv_with_unicode(self, temp_fixtures_dir: Path) -> None:
        """Test CSV with Unicode characters."""
        unicode_data = [
            {"name": "François", "city": "Zürich"},
            {"name": "日本", "city": "東京"},
            {"name": "Москва", "city": "Россия"},
        ]

        csv_file = temp_fixtures_dir / "unicode.csv"
        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=unicode_data[0].keys())
            writer.writeheader()
            writer.writerows(unicode_data)

        result = open_fixture(temp_fixtures_dir, "unicode")
        assert result == unicode_data

    def test_csv_with_embedded_newlines(self, temp_fixtures_dir: Path) -> None:
        """Test CSV with embedded newlines in quoted fields (RFC 4180 compliance)."""
        # This is a critical test - embedded newlines must be preserved
        csv_content = """name,description
Alice,"Line 1
Line 2"
Bob,"Simple description"
Charlie,"Multiple
embedded
newlines"
"""
        csv_file = temp_fixtures_dir / "embedded_newlines.csv"
        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            f.write(csv_content)

        result = open_fixture(temp_fixtures_dir, "embedded_newlines")
        assert len(result) == 3
        assert result[0]["name"] == "Alice"
        assert result[0]["description"] == "Line 1\nLine 2"  # Newline must be preserved
        assert result[1]["description"] == "Simple description"
        assert result[2]["description"] == "Multiple\nembedded\nnewlines"

    def test_csv_with_embedded_newlines_gzip(self, temp_fixtures_dir: Path) -> None:
        """Test gzipped CSV with embedded newlines in quoted fields."""
        csv_content = """name,description
Alice,"Line 1
Line 2"
"""
        gz_file = temp_fixtures_dir / "embedded_newlines_gz.csv.gz"
        with gzip.open(gz_file, "wt", encoding="utf-8", newline="") as f:
            f.write(csv_content)

        result = open_fixture(temp_fixtures_dir, "embedded_newlines_gz")
        assert len(result) == 1
        assert result[0]["description"] == "Line 1\nLine 2"  # Newline must be preserved

    def test_csv_with_embedded_newlines_zip(self, temp_fixtures_dir: Path) -> None:
        """Test zipped CSV with embedded newlines in quoted fields."""
        csv_content = """name,description
Alice,"Line 1
Line 2"
"""
        zip_file = temp_fixtures_dir / "embedded_newlines_zip.csv.zip"
        with zipfile.ZipFile(zip_file, "w", zipfile.ZIP_DEFLATED) as zf:
            zf.writestr("embedded_newlines_zip.csv", csv_content)

        result = open_fixture(temp_fixtures_dir, "embedded_newlines_zip")
        assert len(result) == 1
        assert result[0]["description"] == "Line 1\nLine 2"  # Newline must be preserved

    def test_csv_with_utf8_bom(self, temp_fixtures_dir: Path) -> None:
        """Test CSV with UTF-8 BOM (common in Excel exports)."""
        # UTF-8 BOM is \ufeff at the start of the file
        csv_content = "\ufeffname,value\nAlice,1\nBob,2\n"
        csv_file = temp_fixtures_dir / "bom.csv"
        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            f.write(csv_content)

        result = open_fixture(temp_fixtures_dir, "bom")
        assert len(result) == 2
        # BOM should be stripped, so first key should be "name" not "\ufeffname"
        assert "name" in result[0]
        assert "\ufeffname" not in result[0]
        assert result[0]["name"] == "Alice"

    def test_csv_with_utf8_bom_gzip(self, temp_fixtures_dir: Path) -> None:
        """Test gzipped CSV with UTF-8 BOM."""
        csv_content = "\ufeffname,value\nAlice,1\n"
        # Write as bytes to preserve BOM
        gz_file = temp_fixtures_dir / "bom_gz.csv.gz"
        with gzip.open(gz_file, "wb") as f:
            f.write(csv_content.encode("utf-8"))

        result = open_fixture(temp_fixtures_dir, "bom_gz")
        assert len(result) == 1
        assert "name" in result[0]
        assert "\ufeffname" not in result[0]

    def test_csv_with_utf8_bom_zip(self, temp_fixtures_dir: Path) -> None:
        """Test zipped CSV with UTF-8 BOM."""
        csv_content = "\ufeffname,value\nAlice,1\n"
        zip_file = temp_fixtures_dir / "bom_zip.csv.zip"
        with zipfile.ZipFile(zip_file, "w", zipfile.ZIP_DEFLATED) as zf:
            # Write as bytes to preserve BOM
            zf.writestr("bom_zip.csv", csv_content.encode("utf-8"))

        result = open_fixture(temp_fixtures_dir, "bom_zip")
        assert len(result) == 1
        assert "name" in result[0]
        assert "\ufeffname" not in result[0]

    def test_csv_empty_data_rows(self, temp_fixtures_dir: Path) -> None:
        """Test CSV with headers only (no data rows)."""
        csv_content = "name,email,age\n"
        csv_file = temp_fixtures_dir / "headers_only.csv"
        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            f.write(csv_content)

        result = open_fixture(temp_fixtures_dir, "headers_only")
        assert result == []  # Empty list, no rows

    def test_uppercase_plain_csv(self, temp_fixtures_dir: Path) -> None:
        """Test loading uppercase plain CSV file."""
        csv_content = "name,value\nUPPER,1\n"
        csv_file = temp_fixtures_dir / "UPPERTEST.csv"
        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            f.write(csv_content)

        # Should find UPPERTEST.csv when searching for "uppertest"
        result = open_fixture(temp_fixtures_dir, "uppertest")
        assert len(result) == 1
        assert result[0]["name"] == "UPPER"


class TestOpenFixtureAsync:
    """Test cases for asynchronous open_fixture_async function."""

    @pytest.mark.asyncio
    async def test_open_plain_json_fixture_async(
        self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]"
    ) -> None:
        """Test loading plain JSON fixture asynchronously."""
        result = await open_fixture_async(temp_fixtures_dir, "users")
        assert result == sample_data

    @pytest.mark.asyncio
    async def test_open_gzipped_fixture_async(
        self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]"
    ) -> None:
        """Test loading gzipped JSON fixture asynchronously."""
        result = await open_fixture_async(temp_fixtures_dir, "users_gz")
        assert result == sample_data

    @pytest.mark.asyncio
    async def test_open_zipped_fixture_async(
        self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]"
    ) -> None:
        """Test loading zipped JSON fixture asynchronously."""
        result = await open_fixture_async(temp_fixtures_dir, "users_zip")
        assert result == sample_data

    @pytest.mark.asyncio
    async def test_open_zipped_fixture_multiple_files_async(
        self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]"
    ) -> None:
        """Test loading zipped JSON fixture with multiple files asynchronously."""
        result = await open_fixture_async(temp_fixtures_dir, "users_multi")
        assert result == sample_data

    @pytest.mark.asyncio
    async def test_case_insensitive_support_async(
        self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]"
    ) -> None:
        """Test case-insensitive fixture loading asynchronously."""
        # Create uppercase gzipped file
        uppercase_gz_file = temp_fixtures_dir / "ASYNCCASE.json.gz"
        with gzip.open(uppercase_gz_file, "wt", encoding="utf-8") as f:
            json.dump(sample_data, f)

        # Test that uppercase is found
        result = await open_fixture_async(temp_fixtures_dir, "asynccase")
        assert result == sample_data

    @pytest.mark.asyncio
    async def test_file_format_priority_async(self, temp_fixtures_dir: Path) -> None:
        """Test that plain JSON is preferred over compressed formats in async version."""
        # Create all three formats for the same fixture name
        json_file = temp_fixtures_dir / "priority_async.json"
        gz_file = temp_fixtures_dir / "priority_async.json.gz"
        zip_file = temp_fixtures_dir / "priority_async.json.zip"

        # Different data for each format to test which one is loaded
        plain_data = [{"format": "plain"}]
        gz_data = [{"format": "gzip"}]
        zip_data = [{"format": "zip"}]

        with open(json_file, "w", encoding="utf-8") as f:
            json.dump(plain_data, f)

        with gzip.open(gz_file, "wt", encoding="utf-8") as f:
            json.dump(gz_data, f)

        with zipfile.ZipFile(zip_file, "w", zipfile.ZIP_DEFLATED) as zf:
            zf.writestr("priority_async.json", json.dumps(zip_data))

        # Should load plain JSON first
        result = await open_fixture_async(temp_fixtures_dir, "priority_async")
        assert result == plain_data

    @pytest.mark.asyncio
    async def test_fixture_not_found_async(self, temp_fixtures_dir: Path) -> None:
        """Test FileNotFoundError when fixture doesn't exist in async version."""
        with pytest.raises(FileNotFoundError) as exc_info:
            await open_fixture_async(temp_fixtures_dir, "nonexistent")

        assert "Could not find the nonexistent fixture" in str(exc_info.value)
        assert "(tried .json, .json.gz, .json.zip, .csv, .csv.gz, .csv.zip with case variations)" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_empty_zip_file_async(self, temp_fixtures_dir: Path) -> None:
        """Test error handling for empty zip file in async version."""
        with pytest.raises(ValueError) as exc_info:
            await open_fixture_async(temp_fixtures_dir, "empty")

        assert "No JSON files found in zip archive" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_corrupted_gzip_file_async(self, temp_fixtures_dir: Path) -> None:
        """Test error handling for corrupted gzip file in async version."""
        # Create corrupted gzip file
        corrupted_file = temp_fixtures_dir / "corrupted_async.json.gz"
        with open(corrupted_file, "wb") as f:
            f.write(b"not a gzip file")

        with pytest.raises(OSError) as exc_info:
            await open_fixture_async(temp_fixtures_dir, "corrupted_async")

        assert "Error reading fixture file" in str(exc_info.value)

    @pytest.mark.skip(reason="Import mocking is complex and anyio is required by the project")
    @pytest.mark.asyncio
    async def test_missing_anyio_dependency(self, temp_fixtures_dir: Path) -> None:
        """Test MissingDependencyError when anyio is not available."""
        # Note: This test documents the expected behavior when anyio is not available.
        # In practice, anyio is a required dependency for this project.
        pass

    # Async CSV Tests
    @pytest.mark.asyncio
    async def test_open_plain_csv_fixture_async(
        self, temp_fixtures_dir: Path, sample_csv_data: "list[dict[str, str]]"
    ) -> None:
        """Test loading plain CSV fixture asynchronously."""
        result = await open_fixture_async(temp_fixtures_dir, "users_csv")
        assert result == sample_csv_data

    @pytest.mark.asyncio
    async def test_open_gzipped_csv_fixture_async(
        self, temp_fixtures_dir: Path, sample_csv_data: "list[dict[str, str]]"
    ) -> None:
        """Test loading gzipped CSV fixture asynchronously."""
        result = await open_fixture_async(temp_fixtures_dir, "users_csv_gz")
        assert result == sample_csv_data

    @pytest.mark.asyncio
    async def test_open_zipped_csv_fixture_async(
        self, temp_fixtures_dir: Path, sample_csv_data: "list[dict[str, str]]"
    ) -> None:
        """Test loading zipped CSV fixture asynchronously."""
        result = await open_fixture_async(temp_fixtures_dir, "users_csv_zip")
        assert result == sample_csv_data

    @pytest.mark.asyncio
    async def test_open_zipped_csv_fixture_multiple_files_async(
        self, temp_fixtures_dir: Path, sample_csv_data: "list[dict[str, str]]"
    ) -> None:
        """Test loading zipped CSV fixture with multiple files asynchronously."""
        result = await open_fixture_async(temp_fixtures_dir, "users_csv_multi")
        assert result == sample_csv_data

    @pytest.mark.asyncio
    async def test_json_priority_over_csv_async(
        self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]", sample_csv_data: "list[dict[str, str]]"
    ) -> None:
        """Test that JSON fixtures are loaded before CSV when both exist in async version."""
        json_file = temp_fixtures_dir / "priority_async.json"
        csv_file = temp_fixtures_dir / "priority_async.csv"

        with open(json_file, "w", encoding="utf-8") as f:
            json.dump(sample_data, f)

        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=sample_csv_data[0].keys())
            writer.writeheader()
            writer.writerows(sample_csv_data)

        result = await open_fixture_async(temp_fixtures_dir, "priority_async")
        assert result == sample_data

    @pytest.mark.asyncio
    async def test_empty_csv_zip_file_async(self, temp_fixtures_dir: Path) -> None:
        """Test error handling for empty CSV zip file in async version."""
        with pytest.raises(ValueError) as exc_info:
            await open_fixture_async(temp_fixtures_dir, "empty_csv")
        assert "No CSV files found in zip archive" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_csv_with_embedded_newlines_async(self, temp_fixtures_dir: Path) -> None:
        """Test async CSV loading with embedded newlines in quoted fields."""
        csv_content = """name,description
Alice,"Line 1
Line 2"
"""
        csv_file = temp_fixtures_dir / "embedded_async.csv"
        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            f.write(csv_content)

        result = await open_fixture_async(temp_fixtures_dir, "embedded_async")
        assert len(result) == 1
        assert result[0]["description"] == "Line 1\nLine 2"

    @pytest.mark.asyncio
    async def test_csv_with_embedded_newlines_gzip_async(self, temp_fixtures_dir: Path) -> None:
        """Test async gzipped CSV loading with embedded newlines."""
        csv_content = """name,description
Alice,"Line 1
Line 2"
"""
        gz_file = temp_fixtures_dir / "embedded_gz_async.csv.gz"
        with gzip.open(gz_file, "wt", encoding="utf-8", newline="") as f:
            f.write(csv_content)

        result = await open_fixture_async(temp_fixtures_dir, "embedded_gz_async")
        assert len(result) == 1
        assert result[0]["description"] == "Line 1\nLine 2"

    @pytest.mark.asyncio
    async def test_csv_with_embedded_newlines_zip_async(self, temp_fixtures_dir: Path) -> None:
        """Test async zipped CSV loading with embedded newlines."""
        csv_content = """name,description
Alice,"Line 1
Line 2"
"""
        zip_file = temp_fixtures_dir / "embedded_zip_async.csv.zip"
        with zipfile.ZipFile(zip_file, "w", zipfile.ZIP_DEFLATED) as zf:
            zf.writestr("embedded_zip_async.csv", csv_content)

        result = await open_fixture_async(temp_fixtures_dir, "embedded_zip_async")
        assert len(result) == 1
        assert result[0]["description"] == "Line 1\nLine 2"

    @pytest.mark.asyncio
    async def test_csv_with_utf8_bom_async(self, temp_fixtures_dir: Path) -> None:
        """Test async CSV loading with UTF-8 BOM."""
        csv_content = "\ufeffname,value\nAlice,1\n"
        csv_file = temp_fixtures_dir / "bom_async.csv"
        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            f.write(csv_content)

        result = await open_fixture_async(temp_fixtures_dir, "bom_async")
        assert len(result) == 1
        assert "name" in result[0]
        assert "\ufeffname" not in result[0]

    @pytest.mark.asyncio
    async def test_concurrent_csv_reads(self, temp_fixtures_dir: Path) -> None:
        """Test concurrent async reads of the same CSV fixture."""
        import asyncio

        csv_content = "name,value\nAlice,1\nBob,2\n"
        csv_file = temp_fixtures_dir / "concurrent.csv"
        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            f.write(csv_content)

        # Perform 5 concurrent reads
        results = await asyncio.gather(*(open_fixture_async(temp_fixtures_dir, "concurrent") for _ in range(5)))

        # All results should be identical
        assert len(results) == 5
        for result in results:
            assert len(result) == 2
            assert result[0]["name"] == "Alice"
            assert result[1]["name"] == "Bob"

    @pytest.mark.asyncio
    async def test_uppercase_plain_csv_async(self, temp_fixtures_dir: Path) -> None:
        """Test async loading of uppercase plain CSV file."""
        csv_content = "name,value\nUPPER,1\n"
        csv_file = temp_fixtures_dir / "UPPERASYNC.csv"
        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            f.write(csv_content)

        result = await open_fixture_async(temp_fixtures_dir, "upperasync")
        assert len(result) == 1
        assert result[0]["name"] == "UPPER"


class TestIntegration:
    """Integration tests to ensure compatibility with existing usage patterns."""

    def test_backward_compatibility_sync(self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]") -> None:
        """Test that existing sync code still works as expected."""
        # This mirrors the usage pattern in test_sqlquery_service.py
        fixture_data = open_fixture(temp_fixtures_dir, "users")
        assert fixture_data == sample_data
        assert len(fixture_data) == 3
        assert fixture_data[0]["name"] == "Alice"

    @pytest.mark.asyncio
    async def test_backward_compatibility_async(
        self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]"
    ) -> None:
        """Test that existing async code still works as expected."""
        # This mirrors the usage pattern in test_sqlquery_service.py
        fixture_data = await open_fixture_async(temp_fixtures_dir, "users")
        assert fixture_data == sample_data
        assert len(fixture_data) == 3
        assert fixture_data[0]["name"] == "Alice"

    def test_compression_efficiency(self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]") -> None:
        """Test that compressed fixtures are actually smaller than plain JSON."""
        # Create a larger dataset for meaningful compression
        large_data = sample_data * 100  # Repeat the data 100 times

        # Create plain JSON
        json_file = temp_fixtures_dir / "large.json"
        with open(json_file, "w", encoding="utf-8") as f:
            json.dump(large_data, f, indent=2)

        # Create gzipped version
        gz_file = temp_fixtures_dir / "large.json.gz"
        with gzip.open(gz_file, "wt", encoding="utf-8") as f:
            json.dump(large_data, f, indent=2)

        # Create zipped version
        zip_file = temp_fixtures_dir / "large.json.zip"
        with zipfile.ZipFile(zip_file, "w", zipfile.ZIP_DEFLATED) as zf:
            zf.writestr("large.json", json.dumps(large_data, indent=2))

        # Check file sizes
        json_size = json_file.stat().st_size
        gz_size = gz_file.stat().st_size
        zip_size = zip_file.stat().st_size

        # Compressed versions should be smaller
        assert gz_size < json_size
        assert zip_size < json_size

        # Verify all formats load the same data
        json_data = open_fixture(temp_fixtures_dir, "large")

        # Remove plain JSON to force loading compressed versions
        json_file.unlink()
        gz_data = open_fixture(temp_fixtures_dir, "large")

        gz_file.unlink()
        zip_data = open_fixture(temp_fixtures_dir, "large")

        assert json_data == gz_data == zip_data == large_data

    def test_csv_fixture_integration_sync(self, temp_fixtures_dir: Path) -> None:
        """Test CSV fixture loading in a realistic scenario."""
        # Create a CSV fixture similar to real-world usage
        states_data = [
            {"abbreviation": "AL", "name": "Alabama"},
            {"abbreviation": "AK", "name": "Alaska"},
            {"abbreviation": "AZ", "name": "Arizona"},
        ]

        csv_file = temp_fixtures_dir / "us_state_lookup.csv"
        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=["abbreviation", "name"])
            writer.writeheader()
            writer.writerows(states_data)

        # Load fixture
        fixture_data = open_fixture(temp_fixtures_dir, "us_state_lookup")
        assert fixture_data == states_data
        assert len(fixture_data) == 3
        assert fixture_data[0]["name"] == "Alabama"
        assert fixture_data[0]["abbreviation"] == "AL"

    @pytest.mark.asyncio
    async def test_csv_fixture_integration_async(self, temp_fixtures_dir: Path) -> None:
        """Test CSV fixture loading in a realistic async scenario."""
        states_data = [
            {"abbreviation": "CA", "name": "California"},
            {"abbreviation": "TX", "name": "Texas"},
        ]

        csv_file = temp_fixtures_dir / "states_async.csv"
        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=["abbreviation", "name"])
            writer.writeheader()
            writer.writerows(states_data)

        fixture_data = await open_fixture_async(temp_fixtures_dir, "states_async")
        assert fixture_data == states_data
        assert len(fixture_data) == 2

    def test_mixed_format_directory(
        self, temp_fixtures_dir: Path, sample_data: "list[dict[str, Any]]", sample_csv_data: "list[dict[str, str]]"
    ) -> None:
        """Test loading from directory with mixed JSON and CSV fixtures."""
        # Create JSON fixture
        json_file = temp_fixtures_dir / "data_json.json"
        with open(json_file, "w", encoding="utf-8") as f:
            json.dump(sample_data, f)

        # Create CSV fixture
        csv_file = temp_fixtures_dir / "data_csv.csv"
        with open(csv_file, "w", encoding="utf-8", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=sample_csv_data[0].keys())
            writer.writeheader()
            writer.writerows(sample_csv_data)

        # Both should load correctly
        json_result = open_fixture(temp_fixtures_dir, "data_json")
        csv_result = open_fixture(temp_fixtures_dir, "data_csv")

        assert json_result == sample_data
        assert csv_result == sample_csv_data



"""Pwdlib Hashing Backend."""

from typing import TYPE_CHECKING, Any, Union

from advanced_alchemy.exceptions import MissingDependencyError
from advanced_alchemy.types.password_hash.base import HashingBackend
from advanced_alchemy.typing import PWDLIB_INSTALLED

if TYPE_CHECKING:
    from pwdlib.hashers.base import HasherProtocol
    from sqlalchemy import BinaryExpression, ColumnElement

__all__ = ("PwdlibHasher",)


class PwdlibHasher(HashingBackend):
    """Hashing backend using Pwdlib."""

    def __init__(self, hasher: "HasherProtocol") -> None:
        """Initialize PwdlibBackend.

        Args:
            hasher: The Pwdlib hasher to use for hashing and verification.

        Raises:
            MissingDependencyError: If the ``pwdlib`` package is not installed.
        """
        if not PWDLIB_INSTALLED:
            raise MissingDependencyError(package="pwdlib")
        self.hasher = hasher

    def hash(self, value: "Union[str, bytes]") -> str:
        """Hash the given value using the Pwdlib context.

        Args:
            value: The plain text value to hash. Will be converted to string.

        Returns:
            The hashed string.
        """
        return self.hasher.hash(self._ensure_bytes(value))

    def verify(self, plain: "Union[str, bytes]", hashed: str) -> bool:
        """Verify a plain text value against a hash using the Pwdlib context.

        Args:
            plain: The plain text value to verify. Will be converted to string.
            hashed: The hash to verify against.

        Returns:
            True if the plain text matches the hash, False otherwise.
        """
        try:
            return self.hasher.verify(self._ensure_bytes(plain), hashed)
        except Exception:  # noqa: BLE001
            return False

    def needs_rehash(self, hashed: str) -> bool:
        """Return True if the wrapped hasher reports the stored hash as stale.

        Args:
            hashed: The stored hash string.

        Returns:
            True if the hash should be regenerated; False for an unparsable or foreign hash.
        """
        try:
            return self.hasher.check_needs_rehash(hashed)
        except Exception:  # noqa: BLE001
            return False

    def compare_expression(self, column: "ColumnElement[str]", plain: Any) -> "BinaryExpression[bool]":
        """Direct SQL comparison is not supported for Pwdlib.

        Raises:
            NotImplementedError: Always raised.
        """
        msg = "PwdlibHasher does not support direct SQL comparison."
        raise NotImplementedError(msg)



"""Unit tests for advanced_alchemy.types.json.ORA_JSONB dialect dispatch."""

import sys
from typing import Any
from unittest.mock import MagicMock

import pytest
from sqlalchemy.dialects.oracle import BLOB as ORA_BLOB
from sqlalchemy.engine import Dialect

from advanced_alchemy.types.json import ORA_JSONB


def _make_dialect(server_version_info: Any) -> MagicMock:
    """Build a MagicMock dialect that returns the input descriptor unchanged."""
    dialect = MagicMock(spec=Dialect)
    dialect.name = "oracle"
    dialect.server_version_info = server_version_info
    dialect.type_descriptor.side_effect = lambda t: t
    return dialect


@pytest.fixture
def _no_oracle_json(monkeypatch: pytest.MonkeyPatch) -> None:
    """Simulate SA 2.0.x where sqlalchemy.dialects.oracle has no JSON symbol."""
    import sqlalchemy.dialects.oracle as ora

    monkeypatch.delattr(ora, "JSON", raising=False)
    monkeypatch.setitem(
        sys.modules,
        "sqlalchemy.dialects.oracle",
        ora,
    )


@pytest.fixture
def _with_oracle_json(monkeypatch: pytest.MonkeyPatch) -> type:
    """Simulate SA 2.1+ by injecting a fake JSON class into sqlalchemy.dialects.oracle."""
    import sqlalchemy.dialects.oracle as ora

    class _FakeOracleJSON:
        def __init__(self) -> None:
            self.is_fake_native_oracle_json = True

    monkeypatch.setattr(ora, "JSON", _FakeOracleJSON, raising=False)
    return _FakeOracleJSON


def test_load_dialect_impl_sa20_returns_blob(_no_oracle_json: None) -> None:
    """SA 2.0.x: no native oracle.JSON available, fall back to ORA_BLOB."""
    dialect = _make_dialect(server_version_info=(23, 0))
    result = ORA_JSONB().load_dialect_impl(dialect)
    assert isinstance(result, ORA_BLOB)


def test_load_dialect_impl_sa21_oracle_23c_returns_native_json(_with_oracle_json: type) -> None:
    """SA 2.1+ on Oracle 21c or newer returns native oracle.JSON."""
    dialect = _make_dialect(server_version_info=(23, 0))
    result = ORA_JSONB().load_dialect_impl(dialect)
    assert isinstance(result, _with_oracle_json)


def test_load_dialect_impl_sa21_oracle_21c_returns_native_json(_with_oracle_json: type) -> None:
    """SA 2.1+ on Oracle 21c exactly returns native oracle.JSON."""
    dialect = _make_dialect(server_version_info=(21, 0))
    result = ORA_JSONB().load_dialect_impl(dialect)
    assert isinstance(result, _with_oracle_json)


def test_load_dialect_impl_sa21_oracle_19c_returns_blob(_with_oracle_json: type) -> None:
    """SA 2.1+ on Oracle 19c falls back to ORA_BLOB."""
    dialect = _make_dialect(server_version_info=(19, 0))
    result = ORA_JSONB().load_dialect_impl(dialect)
    assert isinstance(result, ORA_BLOB)


def test_load_dialect_impl_sa21_missing_server_version_returns_blob(_with_oracle_json: type) -> None:
    """SA 2.1+ but server_version_info is None (offline / first-connect) falls back to ORA_BLOB."""
    dialect = _make_dialect(server_version_info=None)
    result = ORA_JSONB().load_dialect_impl(dialect)
    assert isinstance(result, ORA_BLOB)


def test_should_create_constraint_non_oracle_dialect_returns_false() -> None:
    """Non-oracle dialects never get the JSON CHECK constraint."""
    compiler = MagicMock()
    compiler.dialect.name = "postgresql"
    assert ORA_JSONB()._should_create_constraint(compiler) is False


def test_should_create_constraint_sa20_oracle_returns_true(_no_oracle_json: None) -> None:
    """SA 2.0.x oracle still emits the 'is json (strict)' CHECK."""
    compiler = MagicMock()
    compiler.dialect.name = "oracle"
    compiler.dialect.server_version_info = (23, 0)
    assert ORA_JSONB()._should_create_constraint(compiler) is True


def test_should_create_constraint_sa21_oracle_21c_returns_false(_with_oracle_json: type) -> None:
    """SA 2.1+ oracle 21c+ skips CHECK because native JSON validates itself."""
    compiler = MagicMock()
    compiler.dialect.name = "oracle"
    compiler.dialect.server_version_info = (21, 0)
    assert ORA_JSONB()._should_create_constraint(compiler) is False


def test_should_create_constraint_sa21_oracle_19c_returns_true(_with_oracle_json: type) -> None:
    """SA 2.1+ oracle 19c still needs CHECK because we keep ORA_BLOB path."""
    compiler = MagicMock()
    compiler.dialect.name = "oracle"
    compiler.dialect.server_version_info = (19, 0)
    assert ORA_JSONB()._should_create_constraint(compiler) is True


def test_should_create_constraint_sa21_oracle_missing_version_returns_true(_with_oracle_json: type) -> None:
    """SA 2.1+ oracle with no server version info is treated like older Oracle and keeps CHECK."""
    compiler = MagicMock()
    compiler.dialect.name = "oracle"
    compiler.dialect.server_version_info = None
    assert ORA_JSONB()._should_create_constraint(compiler) is True



"""Unit tests for shared dependency-injection utilities."""

from enum import Enum
from typing import Any, cast
from unittest.mock import MagicMock

import pytest

from advanced_alchemy.utils.dependencies import (
    ChoiceField,
    DependencyCache,
    FieldNameType,
    FilterConfig,
    make_hashable,
    normalize_choice_field_types,
    normalize_sort_field,
)
from advanced_alchemy.utils.singleton import SingletonMeta

pytestmark = pytest.mark.unit


class StatusChoice(str, Enum):
    ACTIVE = "active"
    PENDING = "pending"


def test_field_name_type_is_named_tuple_with_default_type_hint() -> None:
    field = FieldNameType(name="id")

    assert field.name == "id"
    assert field.type_hint is str
    assert tuple(field) == ("id", str)


def test_filter_config_allows_empty_config() -> None:
    config: FilterConfig = {}

    assert config == {}


def test_make_hashable_dict_order_invariant() -> None:
    left: dict[str, Any] = {"x": 1, "y": 2}
    right: dict[str, Any] = {"y": 2, "x": 1}

    assert make_hashable(left) == make_hashable(right)


def test_make_hashable_nested_values() -> None:
    config: FilterConfig = cast(
        "FilterConfig",
        {
            "sort_field": ("a", "b"),
            "search_ignore_case": True,
            "in_fields": [FieldNameType("tag")],
            "choice_fields": [ChoiceField("status", ["active", "pending"])],
        },
    )

    assert make_hashable(config) == make_hashable(dict(config))


def test_make_hashable_set_order_invariant() -> None:
    assert make_hashable({1, 2, 3}) == make_hashable({3, 2, 1})


def test_make_hashable_preserves_list_order() -> None:
    assert make_hashable(["name", "id"]) != make_hashable(["id", "name"])


def test_normalize_sort_field_preserves_list_order() -> None:
    assert normalize_sort_field(["name", "id"]) == "name"


def test_normalize_sort_field_sorts_set_values() -> None:
    assert normalize_sort_field({"name", "id"}) == "id"


def test_normalize_choice_field_types_supports_explicit_values() -> None:
    normalized = normalize_choice_field_types([ChoiceField("status", ["active", "pending"])])
    field = next(iter(normalized))

    assert field.name == "status"
    assert "active" in str(field.type_hint)
    assert "pending" in str(field.type_hint)


def test_normalize_choice_field_types_supports_explicit_value_tuples() -> None:
    normalized = normalize_choice_field_types(("status", ("active", "pending")))
    field = next(iter(normalized))

    assert field.name == "status"
    assert "active" in str(field.type_hint)
    assert "pending" in str(field.type_hint)


def test_normalize_choice_field_types_supports_type_hint_tuples() -> None:
    normalized = normalize_choice_field_types(("status", StatusChoice))
    field = next(iter(normalized))

    assert field.name == "status"
    assert field.type_hint is StatusChoice


def test_normalize_choice_field_types_preserves_list_order() -> None:
    normalized = normalize_choice_field_types([("visibility", ["public", "private"]), FieldNameType("status", str)])

    assert [field.name for field in normalized] == ["visibility", "status"]


def test_dependency_cache_singleton() -> None:
    with pytest.MonkeyPatch.context() as monkeypatch:
        monkeypatch.setattr(SingletonMeta, "_instances", {})
        first = DependencyCache()
        second = DependencyCache()

    assert first is second


def test_dependency_cache_add_get_dependencies() -> None:
    with pytest.MonkeyPatch.context() as monkeypatch:
        monkeypatch.setattr(SingletonMeta, "_instances", {})
        cache = DependencyCache()
        dependencies = {"filters": MagicMock()}

        cache.add_dependencies("key", dependencies)

        assert cache.get_dependencies("key") == dependencies
        assert cache.get_dependencies("missing") is None


def test_dependency_cache_get_set_by_config() -> None:
    with pytest.MonkeyPatch.context() as monkeypatch:
        monkeypatch.setattr(SingletonMeta, "_instances", {})
        cache = DependencyCache()
        config: FilterConfig = {"id_field": "uuid"}
        dependencies = {"filters": MagicMock()}

        cache.set(config, dependencies)

        assert cache.get(config) == dependencies



"""Example domain objects for testing."""

from __future__ import annotations

from advanced_alchemy.service import (
    SchemaDumpConfig,
    SQLAlchemyAsyncRepositoryService,
    SQLAlchemySyncRepositoryService,
)
from advanced_alchemy.utils.serialization import ModelDictT, is_dict_with_field, is_dict_without_field, schema_dump
from tests.fixtures.bigint.models import (
    BigIntAuthor,
    BigIntBook,
    BigIntEventLog,
    BigIntItem,
    BigIntModelWithFetchedValue,
    BigIntRule,
    BigIntSecret,
    BigIntSlugBook,
    BigIntTag,
)
from tests.fixtures.bigint.repositories import (
    AuthorAsyncMockRepository,
    AuthorAsyncRepository,
    AuthorSyncMockRepository,
    AuthorSyncRepository,
    BookAsyncMockRepository,
    BookAsyncRepository,
    BookSyncMockRepository,
    BookSyncRepository,
    EventLogAsyncRepository,
    EventLogSyncRepository,
    ItemAsyncMockRepository,
    ItemAsyncRepository,
    ItemSyncMockRepository,
    ItemSyncRepository,
    ModelWithFetchedValueAsyncRepository,
    ModelWithFetchedValueSyncRepository,
    RuleAsyncMockRepository,
    RuleAsyncRepository,
    RuleSyncMockRepository,
    RuleSyncRepository,
    SecretAsyncRepository,
    SecretSyncRepository,
    SlugBookAsyncMockRepository,
    SlugBookAsyncRepository,
    SlugBookSyncMockRepository,
    SlugBookSyncRepository,
    TagAsyncMockRepository,
    TagAsyncRepository,
    TagSyncMockRepository,
    TagSyncRepository,
)


# Services
class SecretAsyncService(SQLAlchemyAsyncRepositoryService[BigIntSecret, SecretAsyncRepository]):
    """Rule repository."""

    repository_type = SecretAsyncRepository


class RuleAsyncService(SQLAlchemyAsyncRepositoryService[BigIntRule, RuleAsyncRepository]):
    """Rule repository."""

    repository_type = RuleAsyncRepository


class RuleAsyncMockService(SQLAlchemyAsyncRepositoryService[BigIntRule, RuleAsyncMockRepository]):
    """Rule repository."""

    repository_type = RuleAsyncMockRepository


class RuleSyncMockService(SQLAlchemySyncRepositoryService[BigIntRule, RuleSyncMockRepository]):
    """Rule repository."""

    repository_type = RuleSyncMockRepository


class AuthorAsyncService(SQLAlchemyAsyncRepositoryService[BigIntAuthor, AuthorAsyncRepository]):
    """Author repository."""

    repository_type = AuthorAsyncRepository


class AuthorAsyncMockService(SQLAlchemyAsyncRepositoryService[BigIntAuthor, AuthorAsyncMockRepository]):
    """Author repository."""

    repository_type = AuthorAsyncMockRepository


class AuthorSyncMockService(SQLAlchemySyncRepositoryService[BigIntAuthor, AuthorSyncMockRepository]):
    """Author repository."""

    repository_type = AuthorSyncMockRepository


class BookAsyncService(SQLAlchemyAsyncRepositoryService[BigIntBook, BookAsyncRepository]):
    """Book repository."""

    repository_type = BookAsyncRepository


class BookAsyncMockService(SQLAlchemyAsyncRepositoryService[BigIntBook, BookAsyncMockRepository]):
    """Book repository."""

    repository_type = BookAsyncMockRepository


class BookSyncMockService(SQLAlchemySyncRepositoryService[BigIntBook, BookSyncMockRepository]):
    """Book repository."""

    repository_type = BookSyncMockRepository


class EventLogAsyncService(SQLAlchemyAsyncRepositoryService[BigIntEventLog, EventLogAsyncRepository]):
    """Event log repository."""

    repository_type = EventLogAsyncRepository


class ModelWithFetchedValueAsyncService(
    SQLAlchemyAsyncRepositoryService[BigIntModelWithFetchedValue, ModelWithFetchedValueAsyncRepository]
):
    """BigIntModelWithFetchedValue repository."""

    repository_type = ModelWithFetchedValueAsyncRepository


class TagAsyncService(SQLAlchemyAsyncRepositoryService[BigIntTag, TagAsyncRepository]):
    """Tag repository."""

    repository_type = TagAsyncRepository


class TagAsyncMockService(SQLAlchemyAsyncRepositoryService[BigIntTag, TagAsyncMockRepository]):
    """Tag repository."""

    repository_type = TagAsyncMockRepository


class TagSyncMockService(SQLAlchemySyncRepositoryService[BigIntTag, TagSyncMockRepository]):
    """Tag repository."""

    repository_type = TagSyncMockRepository


class ItemAsyncService(SQLAlchemyAsyncRepositoryService[BigIntItem, ItemAsyncRepository]):
    """Item repository."""

    repository_type = ItemAsyncRepository


class ItemAsyncMockService(SQLAlchemyAsyncRepositoryService[BigIntItem, ItemAsyncMockRepository]):
    """Item repository."""

    repository_type = ItemAsyncMockRepository


class ItemSyncMockService(SQLAlchemySyncRepositoryService[BigIntItem, ItemSyncMockRepository]):
    """Item repository."""

    repository_type = ItemSyncMockRepository


class RuleSyncService(SQLAlchemySyncRepositoryService[BigIntRule, RuleSyncRepository]):
    """Rule repository."""

    repository_type = RuleSyncRepository


class AuthorSyncService(SQLAlchemySyncRepositoryService[BigIntAuthor, AuthorSyncRepository]):
    """Author repository."""

    repository_type = AuthorSyncRepository


class BookSyncService(SQLAlchemySyncRepositoryService[BigIntBook, BookSyncRepository]):
    """Book repository."""

    repository_type = BookSyncRepository


class EventLogSyncService(SQLAlchemySyncRepositoryService[BigIntEventLog, EventLogSyncRepository]):
    """Event log repository."""

    repository_type = EventLogSyncRepository


class ModelWithFetchedValueSyncService(
    SQLAlchemySyncRepositoryService[BigIntModelWithFetchedValue, ModelWithFetchedValueSyncRepository]
):
    """BigIntModelWithFetchedValue repository."""

    repository_type = ModelWithFetchedValueSyncRepository


class SecretSyncService(SQLAlchemySyncRepositoryService[BigIntSecret, SecretSyncRepository]):
    """Rule repository."""

    repository_type = SecretSyncRepository


class TagSyncService(SQLAlchemySyncRepositoryService[BigIntTag, TagSyncRepository]):
    """Tag repository."""

    repository_type = TagSyncRepository


class ItemSyncService(SQLAlchemySyncRepositoryService[BigIntItem, ItemSyncRepository]):
    """Item repository."""

    repository_type = ItemSyncRepository


# Slug book


class SlugBookAsyncService(SQLAlchemyAsyncRepositoryService[BigIntSlugBook]):
    """Book repository."""

    repository_type = SlugBookAsyncRepository
    match_fields = ["title"]

    async def to_model(
        self,
        data: ModelDictT[BigIntSlugBook],
        operation: str | None = None,
        schema_dump_config: SchemaDumpConfig | None = None,
    ) -> BigIntSlugBook:
        dump_config = schema_dump_config if schema_dump_config is not None else self.schema_dump_config
        data = schema_dump(data, config=dump_config)
        if is_dict_without_field(data, "slug") and operation == "create":
            data["slug"] = await self.repository.get_available_slug(data["title"])
        if is_dict_without_field(data, "slug") and is_dict_with_field(data, "title") and operation == "update":
            data["slug"] = await self.repository.get_available_slug(data["title"])
        return await super().to_model(data, operation, schema_dump_config=dump_config)


class SlugBookSyncService(SQLAlchemySyncRepositoryService[BigIntSlugBook, SlugBookSyncRepository]):
    """Book repository."""

    repository_type = SlugBookSyncRepository
    match_fields = ["title"]

    def to_model(
        self,
        data: ModelDictT[BigIntSlugBook],
        operation: str | None = None,
        schema_dump_config: SchemaDumpConfig | None = None,
    ) -> BigIntSlugBook:
        dump_config = schema_dump_config if schema_dump_config is not None else self.schema_dump_config
        data = schema_dump(data, config=dump_config)
        if is_dict_without_field(data, "slug") and operation == "create":
            data["slug"] = self.repository.get_available_slug(data["title"])
        if is_dict_without_field(data, "slug") and is_dict_with_field(data, "title") and operation == "update":
            data["slug"] = self.repository.get_available_slug(data["title"])
        return super().to_model(data, operation, schema_dump_config=dump_config)


class SlugBookAsyncMockService(SQLAlchemyAsyncRepositoryService[BigIntSlugBook, SlugBookAsyncMockRepository]):
    """Book repository."""

    repository_type = SlugBookAsyncMockRepository
    match_fields = ["title"]

    async def to_model(
        self,
        data: ModelDictT[BigIntSlugBook],
        operation: str | None = None,
        schema_dump_config: SchemaDumpConfig | None = None,
    ) -> BigIntSlugBook:
        dump_config = schema_dump_config if schema_dump_config is not None else self.schema_dump_config
        data = schema_dump(data, config=dump_config)
        if is_dict_without_field(data, "slug") and operation == "create":
            data["slug"] = await self.repository.get_available_slug(data["title"])
        if is_dict_without_field(data, "slug") and is_dict_with_field(data, "title") and operation == "update":
            data["slug"] = await self.repository.get_available_slug(data["title"])
        return await super().to_model(data, operation, schema_dump_config=dump_config)


class SlugBookSyncMockService(SQLAlchemySyncRepositoryService[BigIntSlugBook, SlugBookSyncMockRepository]):
    """Book repository."""

    repository_type = SlugBookSyncMockRepository
    match_fields = ["title"]

    def to_model(
        self,
        data: ModelDictT[BigIntSlugBook],
        operation: str | None = None,
        schema_dump_config: SchemaDumpConfig | None = None,
    ) -> BigIntSlugBook:
        dump_config = schema_dump_config if schema_dump_config is not None else self.schema_dump_config
        data = schema_dump(data, config=dump_config)
        if is_dict_without_field(data, "slug") and operation == "create":
            data["slug"] = self.repository.get_available_slug(data["title"])
        if is_dict_without_field(data, "slug") and is_dict_with_field(data, "title") and operation == "update":
            data["slug"] = self.repository.get_available_slug(data["title"])
        return super().to_model(data, operation, schema_dump_config=dump_config)



from types import ModuleType
from typing import Any, Callable
from unittest.mock import patch

import pytest
from litestar import get
from litestar.config.app import AppConfig
from litestar.plugins import InitPluginProtocol, SerializationPlugin
from litestar.status_codes import HTTP_200_OK
from litestar.testing import RequestFactory, create_test_client
from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column

from advanced_alchemy.base import UUIDAuditBase
from advanced_alchemy.extensions.litestar import (
    SQLAlchemyAsyncConfig,
    SQLAlchemyInitPlugin,
    SQLAlchemySerializationPlugin,
)
from advanced_alchemy.service.pagination import OffsetPagination


def test_serialization_plugin_implements_both_protocols() -> None:
    """It is registered as both a SerializationPlugin and an InitPluginProtocol."""
    plugin = SQLAlchemySerializationPlugin()
    assert isinstance(plugin, SerializationPlugin)
    assert isinstance(plugin, InitPluginProtocol)


def test_init_plugin_preserves_legacy_uuid_type_encoder_registration() -> None:
    """``SQLAlchemyInitPlugin`` keeps the direct-registration UUID encoders."""
    config = SQLAlchemyAsyncConfig(connection_string="sqlite+aiosqlite:///:memory:")
    plugin = SQLAlchemyInitPlugin(config=config)

    class FakeType:
        pass

    class OtherType:
        pass

    def fake_encoder(_: Any) -> str:
        return "fake"

    def other_encoder(_: Any) -> str:
        return "other"

    def override_encoder(_: Any) -> str:
        return "override"

    def fake_predicate(_: Any) -> bool:
        return False

    def fake_decoder(_t: type, _v: Any) -> Any:
        return None

    def user_predicate(_: Any) -> bool:
        return False

    def user_decoder(_t: type, _v: Any) -> Any:
        return None

    with (
        patch(
            "advanced_alchemy.extensions.litestar.plugins.serialization._get_aa_litestar_type_encoders",
            return_value={FakeType: fake_encoder, OtherType: other_encoder},
        ),
        patch(
            "advanced_alchemy.extensions.litestar.plugins.serialization._get_aa_type_decoders",
            return_value=[(fake_predicate, fake_decoder)],
        ),
    ):
        app_config = AppConfig(
            type_encoders={FakeType: override_encoder},
            type_decoders=[(user_predicate, user_decoder)],
        )
        plugin.on_app_init(app_config)

    assert app_config.type_encoders is not None
    assert app_config.type_encoders[FakeType] == override_encoder
    assert app_config.type_encoders[OtherType] == other_encoder
    assert app_config.type_decoders == [
        (user_predicate, user_decoder),
        (fake_predicate, fake_decoder),
    ]


async def test_aa_type_encoders_merging_logic() -> None:
    """Built-in AA encoders are added; user encoders override on collision."""

    class FakeType:
        pass

    def fake_encoder(_: Any) -> str:
        return "fake"

    with patch(
        "advanced_alchemy.extensions.litestar.plugins.serialization._get_aa_type_encoders",
        return_value={FakeType: fake_encoder},
    ):
        plugin = SQLAlchemySerializationPlugin()

        # 1. No user encoders: built-in is added.
        app_config = AppConfig()
        plugin.on_app_init(app_config)
        assert app_config.type_encoders is not None
        assert app_config.type_encoders[FakeType] == fake_encoder

        # 2. User encoder for a different type: both coexist.
        class OtherType:
            pass

        def other_encoder(_: Any) -> str:
            return "other"

        app_config = AppConfig(type_encoders={OtherType: other_encoder})
        plugin.on_app_init(app_config)
        assert app_config.type_encoders is not None
        assert app_config.type_encoders[FakeType] == fake_encoder
        assert app_config.type_encoders[OtherType] == other_encoder

        # 3. User encoder OVERRIDES the AA built-in.
        def override_encoder(_: Any) -> str:
            return "override"

        app_config = AppConfig(type_encoders={FakeType: override_encoder})
        plugin.on_app_init(app_config)
        assert app_config.type_encoders is not None
        assert app_config.type_encoders[FakeType] == override_encoder


async def test_aa_type_decoders_merging_logic() -> None:
    """Built-in AA decoders are added; user decoders take precedence."""

    def fake_predicate(_: Any) -> bool:
        return False

    def fake_decoder(_t: type, _v: Any) -> Any:
        return None

    def user_predicate(_: Any) -> bool:
        return False

    def user_decoder(_t: type, _v: Any) -> Any:
        return None

    with patch(
        "advanced_alchemy.extensions.litestar.plugins.serialization._get_aa_type_decoders",
        return_value=[(fake_predicate, fake_decoder)],
    ):
        plugin = SQLAlchemySerializationPlugin()

        # User decoders must be ordered FIRST because Litestar uses the first
        # matching predicate when resolving decoders.
        app_config = AppConfig(type_decoders=[(user_predicate, user_decoder)])
        plugin.on_app_init(app_config)
        assert app_config.type_decoders is not None
        assert app_config.type_decoders == [
            (user_predicate, user_decoder),
            (fake_predicate, fake_decoder),
        ]


def test_aa_type_decoder_registration_is_idempotent() -> None:
    """``SQLAlchemyInitPlugin`` and ``SQLAlchemySerializationPlugin`` can both run."""
    config = SQLAlchemyAsyncConfig(connection_string="sqlite+aiosqlite:///:memory:")

    def fake_predicate(_: Any) -> bool:
        return False

    def fake_decoder(_t: type, _v: Any) -> Any:
        return None

    decoder = (fake_predicate, fake_decoder)

    with patch(
        "advanced_alchemy.extensions.litestar.plugins.serialization._get_aa_type_decoders",
        return_value=[decoder],
    ):
        app_config = AppConfig()
        SQLAlchemyInitPlugin(config=config).on_app_init(app_config)
        SQLAlchemySerializationPlugin().on_app_init(app_config)

    assert app_config.type_decoders == [decoder]


async def test_real_asyncpg_encoder_integration() -> None:
    """Real asyncpg pgproto.UUID gets registered when asyncpg is installed."""
    pgproto = pytest.importorskip("asyncpg.pgproto.pgproto")

    plugin = SQLAlchemySerializationPlugin()
    app_config = AppConfig()
    plugin.on_app_init(app_config)

    assert app_config.type_encoders is not None
    assert pgproto.UUID in app_config.type_encoders
    assert app_config.type_encoders[pgproto.UUID] is str


async def test_real_uuid_utils_encoder_integration() -> None:
    """Real uuid_utils.UUID gets registered when uuid_utils is installed."""
    uuid_utils = pytest.importorskip("uuid_utils")

    plugin = SQLAlchemySerializationPlugin()
    app_config = AppConfig()
    plugin.on_app_init(app_config)

    assert app_config.type_encoders is not None
    assert uuid_utils.UUID in app_config.type_encoders
    assert app_config.type_encoders[uuid_utils.UUID] is str


async def test_serialization_plugin(
    create_module: Callable[[str], ModuleType],
    request_factory: RequestFactory,
) -> None:
    module = create_module(
        """
from __future__ import annotations

from typing import Dict, List, Set, Tuple, Type, List

from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from litestar import Litestar, get, post
from advanced_alchemy.extensions.litestar import SQLAlchemySerializationPlugin

class Base(DeclarativeBase):
    id: Mapped[int] = mapped_column(primary_key=True)

class A(Base):
    __tablename__ = "a"
    a: Mapped[str]

@post("/a")
def post_handler(data: A) -> A:
    return data

@get("/a")
def get_handler() -> List[A]:
    return [A(id=1, a="test"), A(id=2, a="test2")]

@get("/a/1")
def get_a() -> A:
    return A(id=1, a="test")
""",
    )
    with create_test_client(
        route_handlers=[module.post_handler, module.get_handler, module.get_a],
        plugins=[SQLAlchemySerializationPlugin()],
    ) as client:
        response = client.post("/a", json={"id": 1, "a": "test"})
        assert response.status_code == 201
        assert response.json() == {"id": 1, "a": "test"}
        response = client.get("/a")
        assert response.json() == [{"id": 1, "a": "test"}, {"id": 2, "a": "test2"}]
        response = client.get("/a/1")
        assert response.json() == {"id": 1, "a": "test"}


class User(UUIDAuditBase):
    first_name: Mapped[str] = mapped_column(String(200))


def test_pagination_serialization() -> None:
    users = [User(first_name="ASD"), User(first_name="qwe")]

    @get("/paginated")
    async def paginated_handler() -> OffsetPagination[User]:
        return OffsetPagination[User](items=users, limit=2, offset=0, total=2)

    with create_test_client(paginated_handler, plugins=[SQLAlchemySerializationPlugin()]) as client:
        response = client.get("/paginated")
        assert response.status_code == HTTP_200_OK



from typing import Any, Optional, Union, cast

from sqlalchemy import TypeDecorator

from advanced_alchemy.types.file_object.base import StorageBackend
from advanced_alchemy.types.file_object.file import FileObject
from advanced_alchemy.types.file_object.registry import storages
from advanced_alchemy.types.json import JsonB
from advanced_alchemy.types.mutables import MutableList
from advanced_alchemy.utils.serialization import decode_json

# Define the type hint for the value this TypeDecorator handles
FileObjectOrList = Union[FileObject, list[FileObject], set[FileObject], MutableList[FileObject]]
OptionalFileObjectOrList = Optional[FileObjectOrList]


class StoredObject(TypeDecorator[OptionalFileObjectOrList]):
    """Custom SQLAlchemy type for storing single or multiple file metadata.

    Stores file metadata in JSONB and handles file validation, processing,
    and storage operations through a configured storage backend.
    """

    impl = JsonB
    cache_ok = True

    # Default settings
    multiple: bool
    _raw_backend: Union[str, StorageBackend]
    _resolved_backend: "Optional[StorageBackend]" = None

    @property
    def python_type(self) -> "type[OptionalFileObjectOrList]":
        """Specifies the Python type used, accounting for the `multiple` flag."""
        # This provides a hint to SQLAlchemy and type checkers
        return MutableList[FileObject] if self.multiple else Optional[FileObject]  # type: ignore

    @property
    def backend(self) -> "StorageBackend":
        """Resolves and returns the storage backend instance."""
        # Return cached version if available
        if self._resolved_backend is None:
            self._resolved_backend = (
                storages.get_backend(self._raw_backend) if isinstance(self._raw_backend, str) else self._raw_backend
            )
        return self._resolved_backend

    @property
    def storage_key(self) -> str:
        """Returns the storage key from the resolved backend."""
        return self.backend.key

    def __init__(
        self,
        backend: Union[str, StorageBackend],
        multiple: bool = False,
        *args: "Any",
        **kwargs: "Any",
    ) -> None:
        """Initialize StoredObject type.

        Args:
            backend: Key to retrieve the backend or from the storage registry or storage backend to use.
            multiple: If True, stores a list of files; otherwise, a single file.
            *args: Additional positional arguments for TypeDecorator.
            **kwargs: Additional keyword arguments for TypeDecorator.

        """
        super().__init__(*args, **kwargs)
        self.multiple = multiple
        self._raw_backend = backend

    def __repr__(self) -> str:
        """Return a string representation of the StoredObject."""
        if self.multiple:
            return f"StoredObject(backend='{self.backend.key}', multiple=True)"
        return f"StoredObject(backend='{self.backend.key}')"

    def process_bind_param(
        self,
        value: "Optional[FileObjectOrList]",
        dialect: "Any",
    ) -> "Optional[Union[dict[str, Any], list[dict[str, Any]]]]":
        """Convert FileObject(s) to JSON representation for the database.

        Injects the configured backend into the FileObject before conversion.

        Note: This method expects an already processed FileInfo or its dict representation.
              Use handle_upload() or handle_upload_async() for processing raw uploads.

        Args:
            value: The value to process
            dialect: The SQLAlchemy dialect

        Raises:
            TypeError: If the input value is not a FileObject or a list of FileObjects.

        Returns:
            A dictionary representing the file metadata, or None if the input value is None.
        """
        if value is None:
            return None

        if self.multiple:
            if not isinstance(value, (list, MutableList, set)):
                return [value.to_dict()] if value else []
            return [item.to_dict() for item in value if item]

        if isinstance(value, (list, MutableList, set)):
            msg = f"Expected a single FileObject for multiple=False, got {type(value)}"
            raise TypeError(msg)

        return value.to_dict() if value else None

    def process_result_value(
        self, value: "Optional[Union[bytes, str, dict[str, Any], list[dict[str, Any]]]]", dialect: "Any"
    ) -> "Optional[FileObjectOrList]":
        """Convert database JSON back to FileObject or MutableList[FileObject].

        Args:
            value: The value to process
            dialect: The SQLAlchemy dialect

        Raises:
            TypeError: If the input value is not a list of dicts.

        Returns:
            FileObject or MutableList[FileObject] or None.
        """
        if value is None:
            return None

        if self.multiple:
            if isinstance(value, dict):
                # If the DB returns a single dict, wrap it in a list
                value = [value]
            elif isinstance(value, (str, bytes)):
                # Decode JSON string or bytes to dict
                value = [cast("dict[str, Any]", decode_json(value))]
            return MutableList[FileObject]([FileObject(**v) for v in value if v])  # pyright: ignore
        if isinstance(value, list):
            msg = f"Expected dict from DB for multiple=False, got {type(value)}"
            raise TypeError(msg)
        if isinstance(value, (bytes, str)):
            value = cast("dict[str,Any]", decode_json(value))
        return FileObject(**value)



"""Shared utilities for framework dependency-injection wiring.

These primitives are framework-agnostic and reused by framework
``advanced_alchemy.extensions`` provider modules.
"""

from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from typing import Any, Literal, NamedTuple, Optional, Union, cast
from uuid import UUID

from typing_extensions import NotRequired, TypeAlias, TypedDict, TypeGuard

from advanced_alchemy.utils.singleton import SingletonMeta

__all__ = (
    "ChoiceField",
    "DependencyCache",
    "FieldNameType",
    "FilterConfig",
    "make_hashable",
    "normalize_choice_field_types",
    "normalize_field_name_types",
    "normalize_sort_field",
)

HashableValue = Union[str, int, float, bool, None]
HashableType = Union[HashableValue, tuple[Any, ...], tuple[tuple[str, Any], ...], tuple[HashableValue, ...]]
SortOrder = Literal["asc", "desc"]
SortField = Union[str, set[str], list[str]]
FIELD_TUPLE_LENGTH = 2


class FieldNameType(NamedTuple):
    """Type for field name and associated type information.

    This allows callers to specify both the field name and the expected type
    for generated filter values.
    """

    name: str
    """Name of the field to filter on."""
    type_hint: Any = str
    """Type of the filter value. Defaults to ``str``."""


@dataclass(frozen=True)
class ChoiceField:
    """Field name and explicit values for generated choice filters."""

    name: str
    """Name of the field to filter on."""
    choices: tuple[Any, ...]
    """Allowed values for the generated choice filter."""

    def __init__(self, name: str, choices: Iterable[Any]) -> None:
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "choices", tuple(choices))


FieldDefinition: TypeAlias = Union[str, FieldNameType]
FieldNameConfig: TypeAlias = Union[FieldDefinition, set[FieldDefinition], list[FieldDefinition]]
ChoiceFieldTuple: TypeAlias = tuple[str, Iterable[Any]]
FieldNameTypeTuple: TypeAlias = tuple[str, Any]
ChoiceFieldDefinition: TypeAlias = Union[str, FieldNameType, ChoiceField, FieldNameTypeTuple]
ChoiceFieldConfig: TypeAlias = Union[
    ChoiceFieldDefinition, set[ChoiceFieldDefinition], list[ChoiceFieldDefinition], tuple[ChoiceFieldDefinition, ...]
]


class FilterConfig(TypedDict):
    """Configuration for generating dynamic filters."""

    id_filter: NotRequired[type[Union[UUID, int, str]]]
    """Enable an id filter using the supplied type."""
    id_field: NotRequired[str]
    """Model field storing the primary key or identifier."""
    sort_field: NotRequired[SortField]
    """Default field or fields to use for sorting."""
    sort_order: NotRequired[SortOrder]
    """Default sort order."""
    pagination_type: NotRequired[Literal["limit_offset"]]
    """Pagination mode to enable."""
    pagination_size: NotRequired[int]
    """Default pagination size."""
    search: NotRequired[Union[str, set[str], list[str]]]
    """Fields to enable search on."""
    search_ignore_case: NotRequired[bool]
    """Whether search should be case-insensitive."""
    created_at: NotRequired[bool]
    """Enable created-at range filtering."""
    updated_at: NotRequired[bool]
    """Enable updated-at range filtering."""
    not_in_fields: NotRequired[FieldNameConfig]
    """Fields that support not-in collection filters."""
    in_fields: NotRequired[FieldNameConfig]
    """Fields that support in-collection filters."""
    boolean_fields: NotRequired[FieldNameConfig]
    """Fields that support boolean filters."""
    choice_fields: NotRequired[ChoiceFieldConfig]
    """Fields that support choices filters."""


class DependencyCache(metaclass=SingletonMeta):
    """Singleton cache for dynamically generated dependency providers."""

    def __init__(self) -> None:
        self.dependencies: dict[HashableType, Any] = {}

    def add_dependencies(self, key: HashableType, dependencies: Any) -> None:
        """Cache dependency providers by a precomputed key.

        Args:
            key: Precomputed cache key.
            dependencies: Dependency provider object to cache.
        """
        self.dependencies[key] = dependencies

    def get_dependencies(self, key: HashableType) -> Optional[Any]:
        """Retrieve dependency providers by a precomputed key.

        Args:
            key: Precomputed cache key.

        Returns:
            Cached dependency providers, if present.
        """
        return self.dependencies.get(key)

    def set(self, config: FilterConfig, dependencies: Any) -> None:
        """Cache dependency providers by filter config.

        Args:
            config: Filter configuration used to derive a cache key.
            dependencies: Dependency provider object to cache.
        """
        self.dependencies[make_hashable(config)] = dependencies

    def get(self, config: FilterConfig) -> Optional[Any]:
        """Retrieve dependency providers by filter config.

        Args:
            config: Filter configuration used to derive a cache key.

        Returns:
            Cached dependency providers, if present.
        """
        return self.dependencies.get(make_hashable(config))


def make_hashable(value: Any) -> HashableType:
    """Convert a possibly nested value into a hashable representation.

    Args:
        value: Value that needs to be made hashable.

    Returns:
        A hashable representation of ``value``.
    """
    if isinstance(value, Mapping):
        mapping = cast("Mapping[Any, Any]", value)  # type: ignore[redundant-cast]
        items = [(str(key), make_hashable(mapping[key])) for key in sorted(mapping.keys(), key=str)]
        return tuple(items)
    if isinstance(value, list):
        values = cast("Iterable[Any]", value)
        return tuple(make_hashable(item) for item in values)
    if isinstance(value, set):
        values = cast("Iterable[Any]", value)
        hashable_items = [make_hashable(item) for item in values]
        filtered_items = [item for item in hashable_items if item is not None]
        return tuple(sorted(filtered_items, key=str))
    if isinstance(value, tuple):
        values = cast("tuple[Any, ...]", value)  # type: ignore[redundant-cast]
        return tuple(make_hashable(item) for item in values)
    if isinstance(value, (str, int, float, bool, type(None))):
        return value
    return str(value)


def normalize_field_name_types(field_definitions: FieldNameConfig) -> set[FieldNameType]:
    """Normalize field-name filter config to ``FieldNameType`` values."""
    raw_fields = {field_definitions} if isinstance(field_definitions, (str, FieldNameType)) else set(field_definitions)
    return {
        FieldNameType(name=field_definition, type_hint=str) if isinstance(field_definition, str) else field_definition
        for field_definition in raw_fields
    }


def _literal_type(choices: tuple[Any, ...]) -> Any:
    """Build a runtime ``Literal`` type from configured choices."""
    return cast("Any", Literal).__getitem__(choices)


def _is_choice_field_tuple(value: Any) -> TypeGuard[ChoiceFieldTuple]:
    """Return whether ``value`` is a ``(field_name, choices)`` config tuple."""
    if not isinstance(value, tuple) or isinstance(value, FieldNameType):
        return False
    tuple_value = cast("tuple[Any, ...]", value)  # type: ignore[redundant-cast]
    return (
        len(tuple_value) == FIELD_TUPLE_LENGTH
        and isinstance(tuple_value[0], str)
        and not isinstance(tuple_value[1], type)
        and isinstance(tuple_value[1], Iterable)
        and not isinstance(tuple_value[1], (str, bytes))
    )


def _is_field_name_type_tuple(value: Any) -> TypeGuard[FieldNameTypeTuple]:
    """Return whether ``value`` is a ``(field_name, type_hint)`` config tuple."""
    if not isinstance(value, tuple) or isinstance(value, FieldNameType):
        return False
    tuple_value = cast("tuple[Any, ...]", value)  # type: ignore[redundant-cast]
    return (
        len(tuple_value) == FIELD_TUPLE_LENGTH
        and isinstance(tuple_value[0], str)
        and not _is_choice_field_tuple(value)
        and not isinstance(tuple_value[1], (str, bytes))
    )


def _field_type_from_choices(name: str, choices: Iterable[Any]) -> FieldNameType:
    """Build a field definition from explicit choice values."""
    choices_tuple = tuple(choices)
    if not choices_tuple:
        msg = "ChoiceField choices must not be empty"
        raise ValueError(msg)
    return FieldNameType(name=name, type_hint=_literal_type(choices_tuple))


def normalize_choice_field_types(field_definitions: ChoiceFieldConfig) -> tuple[FieldNameType, ...]:
    """Normalize choice filter config to ``FieldNameType`` values."""
    raw_fields: tuple[ChoiceFieldDefinition, ...]
    if (
        isinstance(field_definitions, (str, FieldNameType, ChoiceField))
        or _is_choice_field_tuple(field_definitions)
        or _is_field_name_type_tuple(field_definitions)
    ):
        raw_fields = (field_definitions,)
    elif isinstance(field_definitions, set):
        raw_fields = tuple(sorted(field_definitions, key=str))
    else:
        raw_fields = tuple(field_definitions)
    normalized_fields: dict[str, FieldNameType] = {}
    for field_definition in raw_fields:
        if isinstance(field_definition, str):
            normalized_field = FieldNameType(name=field_definition, type_hint=str)
        elif isinstance(field_definition, FieldNameType):
            normalized_field = field_definition
        elif isinstance(field_definition, ChoiceField):
            normalized_field = _field_type_from_choices(field_definition.name, field_definition.choices)
        elif _is_choice_field_tuple(field_definition):
            normalized_field = _field_type_from_choices(field_definition[0], field_definition[1])
        elif _is_field_name_type_tuple(field_definition):
            normalized_field = FieldNameType(name=field_definition[0], type_hint=field_definition[1])
        else:
            msg = "Choice field tuple must be (field_name, type_hint) or (field_name, choices)"
            raise ValueError(msg)
        normalized_fields.setdefault(normalized_field.name, normalized_field)
    return tuple(normalized_fields.values())


def normalize_sort_field(sort_field: SortField) -> str:
    """Return a scalar default field for a configured sort field value."""
    if isinstance(sort_field, str):
        return sort_field
    if isinstance(sort_field, list):
        return sort_field[0]
    return sorted(sort_field)[0]



"""Generic unified storage protocol compatible with multiple backend implementations."""

import mimetypes
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

from sqlalchemy.ext.mutable import MutableList
from typing_extensions import TypeAlias

from advanced_alchemy.exceptions import MissingDependencyError
from advanced_alchemy.types.file_object._typing import PYDANTIC_INSTALLED, GetCoreSchemaHandler, core_schema
from advanced_alchemy.types.file_object.base import AsyncDataLike, DataLike, StorageBackend
from advanced_alchemy.types.file_object.registry import storages

if TYPE_CHECKING:
    from advanced_alchemy.types.file_object.base import PathLike


class FileObject:
    """Represents file metadata during processing using a dataclass structure.

    This class provides a unified interface for handling file metadata and operations
    across different storage backends.

    Content or a source path can optionally be provided during initialization via kwargs, store it internally, and add save/save_async methods to persist this pending data using the configured backend.
    """

    __slots__ = (
        "_checksum",
        "_content_type",
        "_etag",
        "_filename",
        "_last_modified",
        "_metadata",
        "_pending_source_content",
        "_pending_source_path",
        "_raw_backend",
        "_resolved_backend",
        "_size",
        "_to_filename",
        "_version_id",
    )

    def __init__(
        self,
        backend: "Union[str, StorageBackend]",
        filename: str,
        to_filename: Optional[str] = None,
        content_type: Optional[str] = None,
        size: Optional[int] = None,
        last_modified: Optional[float] = None,
        checksum: Optional[str] = None,
        etag: Optional[str] = None,
        version_id: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
        source_path: "Optional[PathLike]" = None,
        content: "Optional[Union[DataLike, AsyncDataLike]]" = None,
    ) -> None:
        """Perform post-initialization validation and setup.

        Handles default path, content type guessing, backend protocol inference,
        and processing of 'content' or 'source_path' from extra kwargs.

        Raises:
            ValueError: If filename is not provided, size is negative, backend/protocol mismatch,
                        or both 'content' and 'source_path' are provided.
        """
        self._size = size
        self._last_modified = last_modified
        self._checksum = checksum
        self._etag = etag
        self._version_id = version_id
        self._metadata = metadata or {}
        self._filename = filename
        self._content_type = content_type
        self._to_filename = to_filename
        self._resolved_backend: Optional[StorageBackend] = backend if isinstance(backend, StorageBackend) else None
        self._raw_backend = backend
        self._pending_source_path = Path(source_path) if source_path is not None else None
        self._pending_source_content = content
        if self._pending_source_content is not None and self._pending_source_path is not None:
            msg = "Cannot provide both 'source_content' and 'source_path' during initialization."
            raise ValueError(msg)

    def __repr__(self) -> str:
        """Return a string representation of the FileObject."""
        return f"FileObject(filename={self.path}, backend={self.backend.key}, size={self.size}, content_type={self.content_type}, etag={self.etag}, last_modified={self.last_modified}, version_id={self.version_id})"

    def __eq__(self, other: object) -> bool:
        """Check equality based on filename and backend key.

        Args:
            other: The object to compare with.


        Returns:
            bool: True if the objects are equal, False otherwise.

        """
        if not isinstance(other, FileObject):
            return False
        return self.path == other.path and self.backend.key == other.backend.key

    def __hash__(self) -> int:
        """Return a hash based on filename and backend key."""
        return hash((self.path, self.backend.key))

    @property
    def backend(self) -> "StorageBackend":
        if self._resolved_backend is None:
            self._resolved_backend = (
                storages.get_backend(self._raw_backend) if isinstance(self._raw_backend, str) else self._raw_backend
            )
        return self._resolved_backend

    @property
    def filename(self) -> str:
        return self.path

    @property
    def content_type(self) -> str:
        if self._content_type is None:
            guessed_type, _ = mimetypes.guess_type(self._filename)
            self._content_type = guessed_type or "application/octet-stream"
        return self._content_type

    @property
    def protocol(self) -> str:
        return self.backend.protocol if self.backend else "file"

    @property
    def path(self) -> str:
        return self._to_filename or self._filename

    @property
    def has_pending_data(self) -> bool:
        return bool(self._pending_source_content or self._pending_source_path)

    @property
    def metadata(self) -> dict[str, Any]:
        return self._metadata

    @metadata.setter
    def metadata(self, value: dict[str, Any]) -> None:
        self._metadata = value

    @property
    def size(self) -> "Optional[int]":
        return self._size

    @size.setter
    def size(self, value: int) -> None:
        self._size = value

    @property
    def last_modified(self) -> "Optional[float]":
        return self._last_modified

    @last_modified.setter
    def last_modified(self, value: float) -> None:
        self._last_modified = value

    @property
    def checksum(self) -> "Optional[str]":
        return self._checksum

    @checksum.setter
    def checksum(self, value: str) -> None:
        self._checksum = value

    @property
    def etag(self) -> "Optional[str]":
        return self._etag

    @etag.setter
    def etag(self, value: str) -> None:
        self._etag = value

    @property
    def version_id(self) -> "Optional[str]":
        return self._version_id

    @version_id.setter
    def version_id(self, value: str) -> None:
        self._version_id = value

    def update_metadata(self, metadata: "dict[str, Any]") -> None:
        """Update the file metadata.

        Args:
            metadata: New metadata to merge with existing metadata.
        """
        self.metadata.update(metadata)

    def to_dict(self) -> "dict[str, Any]":
        """Convert FileObject to a dictionary for storage or serialization.

        Note: The 'backend' attribute is intentionally excluded as it's often
              not serializable or relevant for storage representations.
              The 'extra' dict is included.

        Returns:
            dict[str, Any]: A dictionary representation of the file information.
        """
        # Use dataclasses.asdict and filter out the backend
        return {
            "filename": self.path,
            "content_type": self.content_type,
            "size": self.size,
            "last_modified": self.last_modified,
            "checksum": self.checksum,
            "etag": self.etag,
            "version_id": self.version_id,
            "metadata": self.metadata,
            "backend": self.backend.key,
        }

    def get_content(self, *, options: "Optional[dict[str, Any]]" = None) -> bytes:
        """Get the file content from the storage backend.

        Args:
            options: Optional backend-specific options.

        Returns:
            bytes: The file content.
        """
        return self.backend.get_content(self.path, options=options)

    async def get_content_async(self, *, options: "Optional[dict[str, Any]]" = None) -> bytes:
        """Get the file content from the storage backend asynchronously.

        Args:
            options: Optional backend-specific options.

        Returns:
            bytes: The file content.
        """
        return await self.backend.get_content_async(self.path, options=options)

    def sign(
        self,
        *,
        expires_in: "Optional[int]" = None,
        for_upload: bool = False,
    ) -> str:
        """Generate a signed URL for the file.

        Args:
            expires_in: Optional expiration time in seconds.
            for_upload: Whether the URL is for upload.

        Raises:
            RuntimeError: If no signed URL is generated.

        Returns:
            str: The signed URL.
        """
        result = self.backend.sign(self.path, expires_in=expires_in, for_upload=for_upload)
        if isinstance(result, list):
            if not result:
                msg = "No signed URL generated"
                raise RuntimeError(msg)
            return result[0]
        return result

    async def sign_async(
        self,
        *,
        expires_in: "Optional[int]" = None,
        for_upload: bool = False,
    ) -> str:
        """Generate a signed URL for the file asynchronously.

        Args:
            expires_in: Optional expiration time in seconds.
            for_upload: Whether the URL is for upload.

        Returns:
            str: The signed URL.

        Raises:
            RuntimeError: If no signed URL is generated.
        """
        result = await self.backend.sign_async(self.path, expires_in=expires_in, for_upload=for_upload)
        if isinstance(result, list):
            if not result:
                msg = "No signed URL generated"
                raise RuntimeError(msg)
            return result[0]
        return result

    def delete(self) -> None:
        """Delete the file from storage.

        Raises:
            RuntimeError: If no backend is configured or path is missing.
        """
        if not self.backend:
            msg = "No storage backend configured"
            raise RuntimeError(msg)
        self.backend.delete_object(self.path)

    async def delete_async(self) -> None:
        """Delete the file from storage asynchronously."""
        await self.backend.delete_object_async(self.path)

    def save(
        self,
        data: Optional[DataLike] = None,
        *,
        use_multipart: Optional[bool] = None,
        chunk_size: int = 5 * 1024 * 1024,
        max_concurrency: int = 12,
    ) -> "FileObject":
        """Save data to the storage backend using this FileObject's metadata.

        If `data` is provided, it is used directly.
        If `data` is None, checks internal source_content or source_path.
        Clears pending attributes after successful save.

        Args:
            data: Optional data to save (bytes, iterator, file-like, Path). If None,
                  internal pending data is used.
            use_multipart: Passed to the backend's save method.
            chunk_size: Passed to the backend's save method.
            max_concurrency: Passed to the backend's save method.

        Returns:
            The updated FileObject instance returned by the backend.

        Raises:
            TypeError: If trying to save async data synchronously.
        """

        if data is None and self._pending_source_content is not None:
            data = self._pending_source_content  # type: ignore[assignment]
        elif data is None and self._pending_source_path is not None:
            data = self._pending_source_path

        if data is None:
            msg = "No data provided and no pending content/path found to save."
            raise TypeError(msg)

        # The backend's save method is expected to update the FileObject instance in-place
        # and return the updated instance.
        updated_self = self.backend.save_object(
            file_object=self,
            data=data,
            use_multipart=use_multipart,
            chunk_size=chunk_size,
            max_concurrency=max_concurrency,
        )

        # Clear pending attributes after successful save
        self._pending_source_content = None
        self._pending_source_path = None

        return updated_self

    async def save_async(
        self,
        data: Optional[AsyncDataLike] = None,
        *,
        use_multipart: Optional[bool] = None,
        chunk_size: int = 5 * 1024 * 1024,
        max_concurrency: int = 12,
    ) -> "FileObject":
        """Save data to the storage backend asynchronously.

        If `data` is provided, it is used directly.
        If `data` is None, checks internal source_content or source_path.
        Clears pending attributes after successful save.
        Uses asyncio.to_thread for reading source_path if backend doesn't handle Path directly.

        Args:
            data: Optional data to save (bytes, async iterator, file-like, Path, etc.).
                  If None, internal pending data is used.
            use_multipart: Passed to the backend's async save method.
            chunk_size: Passed to the backend's async save method.
            max_concurrency: Passed to the backend's async save method.

        Returns:
            The updated FileObject instance returned by the backend.

        Raises:
            TypeError: If trying to save sync data asynchronously.
        """

        if data is None and self._pending_source_content is not None:
            data = self._pending_source_content
        elif data is None and self._pending_source_path is not None:
            data = self._pending_source_path

        if data is None:
            msg = "No data provided and no pending content/path found to save."
            raise TypeError(msg)

        # Backend's async save method updates the FileObject instance
        updated_self = await self.backend.save_object_async(
            file_object=self,
            data=data,  # Pass the determined data source
            use_multipart=use_multipart,
            chunk_size=chunk_size,
            max_concurrency=max_concurrency,
        )

        # Clear pending attributes after successful save
        self._pending_source_content = None
        self._pending_source_path = None

        return updated_self

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,
        handler: "GetCoreSchemaHandler",  # Use imported GetCoreSchemaHandler
    ) -> "core_schema.CoreSchema":  # Use imported core_schema
        """Get the Pydantic core schema for FileObject.

        This method defines how Pydantic should validate and serialize FileObject instances.
        It creates a schema that validates dictionaries with the required fields and
        converts them to FileObject instances.

        Raises:
            MissingDependencyError: If Pydantic is not installed when this method is called.

        Args:
            source_type: The source type (FileObject)
            handler: The Pydantic schema handler

        Returns:
            A Pydantic core schema for FileObject
        """
        if not PYDANTIC_INSTALLED:
            raise MissingDependencyError(package="pydantic")

        def validate_from_dict(data: dict[str, Any]) -> "FileObject":
            # We expect a dictionary derived from to_dict()
            # We need to resolve the backend string back to an instance if needed
            backend_input = data.get("backend")
            if backend_input is None:
                msg = "backend is required"
                raise TypeError(msg)
            key = backend_input if isinstance(backend_input, str) else backend_input.key
            return cls(
                backend=key,
                filename=data["filename"],
                to_filename=data.get("to_filename"),
                content_type=data.get("content_type"),
                size=data.get("size"),
                last_modified=data.get("last_modified"),
                checksum=data.get("checksum"),
                etag=data.get("etag"),
                version_id=data.get("version_id"),
                metadata=data.get("metadata"),
            )

        typed_dict_schema = core_schema.typed_dict_schema(
            {
                "filename": core_schema.typed_dict_field(core_schema.str_schema()),
                "backend": core_schema.typed_dict_field(core_schema.str_schema()),
                "to_filename": core_schema.typed_dict_field(core_schema.str_schema(), required=False),
                "content_type": core_schema.typed_dict_field(core_schema.str_schema(), required=False),
                "size": core_schema.typed_dict_field(core_schema.int_schema(), required=False),
                "last_modified": core_schema.typed_dict_field(core_schema.float_schema(), required=False),
                "checksum": core_schema.typed_dict_field(core_schema.str_schema(), required=False),
                "etag": core_schema.typed_dict_field(core_schema.str_schema(), required=False),
                "version_id": core_schema.typed_dict_field(core_schema.str_schema(), required=False),
                "metadata": core_schema.typed_dict_field(
                    core_schema.nullable_schema(
                        core_schema.dict_schema(core_schema.str_schema(), core_schema.any_schema())
                    ),
                    required=False,
                ),
            }
        )

        validation_schema = core_schema.union_schema(
            [
                core_schema.is_instance_schema(cls),
                core_schema.chain_schema(
                    [
                        typed_dict_schema,
                        core_schema.no_info_plain_validator_function(validate_from_dict),
                    ]
                ),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=validation_schema,
            python_schema=validation_schema,
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: instance.to_dict(),  # pyright: ignore
                info_arg=False,
                return_schema=typed_dict_schema,
            ),  # pyright: ignore
        )


FileObjectList: TypeAlias = MutableList[FileObject]
